#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
xmrwallet.py - MoneroWalletOps class
"""

import re,time,json,atexit
from collections import namedtuple
from pathlib import Path

from .objmethods import MMGenObject,HiliteStr,InitErrors
from .obj import CoinTxID,Int
from .color import red,yellow,green,blue,cyan,pink,orange,purple,gray
from .util import (
	msg,
	msg_r,
	gmsg,
	bmsg,
	ymsg,
	rmsg,
	gmsg_r,
	pp_msg,
	die,
	fmt,
	suf,
	async_run,
	make_timestr,
	make_chksum_N,
	capfirst,
	list_gen,
	fmt_dict
)
from .fileutil import get_data_from_file
from .seed import SeedID
from .protocol import init_proto
from .proto.btc.common import b58a
from .addr import CoinAddr,AddrIdx
from .addrlist import KeyAddrList,ViewKeyAddrList,AddrIdxList
from .rpc import json_encoder
from .proto.xmr.rpc import MoneroRPCClient,MoneroWalletRPCClient
from .proto.xmr.daemon import MoneroWalletDaemon
from .ui import keypress_confirm
from .autosign import Autosign

xmrwallet_uargs = namedtuple('xmrwallet_uargs',[
	'infile',
	'wallets',
	'spec',
])

xmrwallet_uarg_info = (
	lambda e,hp: {
		'daemon':          e('HOST:PORT', hp),
		'tx_relay_daemon': e('HOST:PORT[:PROXY_IP:PROXY_PORT]',     rf'({hp})(?::({hp}))?'),
		'newaddr_spec':    e('WALLET[:ACCOUNT][,"label text"]',     r'(\d+)(?::(\d+))?(?:,(.*))?'),
		'transfer_spec':   e('SOURCE:ACCOUNT:ADDRESS,AMOUNT',       rf'(\d+):(\d+):([{b58a}]+),([0-9.]+)'),
		'sweep_spec':      e('SOURCE:ACCOUNT[,DEST[:ACCOUNT]]',     r'(\d+):(\d+)(?:,(\d+)(?::(\d+))?)?'),
		'label_spec':      e('WALLET:ACCOUNT:ADDRESS,"label text"', r'(\d+):(\d+):(\d+),(.*)'),
	})(
		namedtuple('uarg_info_entry',['annot','pat']),
		r'(?:[^:]+):(?:\d+)'
	)

def get_autosign_obj(cfg):
	from .cfg import Config
	return Autosign(
		Config({
			'mountpoint': cfg.autosign_mountpoint,
			'test_suite': cfg.test_suite,
			'test_suite_root_pfx': cfg.test_suite_root_pfx,
			'coins': 'xmr',
			'online': not cfg.offline,
		})
	)

# required to squelch pylint:
def fmt_amt(amt):
	return str(amt)

def hl_amt(amt):
	return str(amt)

tx_priorities = {
	1: 'low',
	2: 'normal',
	3: 'high',
	4: 'highest'
}

def gen_acct_addr_info(self, wallet_data, account, indent=''):
	fs = indent + '{I:<3} {A} {U} {B} {L}'
	addrs_data = wallet_data.addrs_data[account]['addresses']

	for d in addrs_data:
		d['unlocked_balance'] = 0

	if 'per_subaddress' in wallet_data.bals_data:
		for d in wallet_data.bals_data['per_subaddress']:
			if d['account_index'] == account:
				addrs_data[d['address_index']]['unlocked_balance'] = d['unlocked_balance']

	yield fs.format(
		I = '',
		A = 'Address'.ljust(addr_width),
		U = 'Used'.ljust(5),
		B = '  Unlocked Balance',
		L = 'Label')

	for addr in addrs_data:
		ca = CoinAddr(self.proto, addr['address'])
		bal = addr['unlocked_balance']
		if self.cfg.skip_empty_addresses and addr['used'] and not bal:
			continue
		yield fs.format(
			I = addr['address_index'],
			A = ca.hl() if self.cfg.full_address else ca.fmt(color=True, width=addr_width),
			U = (red('True ') if addr['used'] else green('False')),
			B = fmt_amt(bal),
			L = pink(addr['label']))

class XMRWalletAddrSpec(HiliteStr,InitErrors,MMGenObject):
	color = 'cyan'
	width = 0
	trunc_ok = False
	min_len = 5  # 1:0:0
	max_len = 14 # 9999:9999:9999
	def __new__(cls,arg1,arg2=None,arg3=None):
		if isinstance(arg1,cls):
			return arg1

		try:
			if isinstance(arg1,str):
				me = str.__new__(cls,arg1)
				m = re.fullmatch( '({n}):({n}):({n}|None)'.format(n=r'[0-9]{1,4}'), arg1 )
				assert m is not None, f'{arg1!r}: invalid XMRWalletAddrSpec'
				for e in m.groups():
					if len(e) != 1 and e[0] == '0':
						die(2,f'{e}: leading zeroes not permitted in XMRWalletAddrSpec element')
				me.wallet = AddrIdx(m[1])
				me.account = int(m[2])
				me.account_address = None if m[3] == 'None' else int(m[3])
			else:
				me = str.__new__(cls,f'{arg1}:{arg2}:{arg3}')
				for arg in [arg1,arg2] + ([] if arg3 is None else [arg3]):
					assert isinstance(arg,int), f'{arg}: XMRWalletAddrSpec component not of type int'
					assert arg is None or arg <= 9999, f'{arg}: XMRWalletAddrSpec component greater than 9999'
				me.wallet = AddrIdx(arg1)
				me.account = arg2
				me.account_address = arg3
			return me
		except Exception as e:
			return cls.init_fail(e,me)

def is_xmr_tx_file(cfg,fn):
	try:
		MoneroMMGenTX.Completed(cfg,fn)
		return True
	except Exception as e:
		if not 'MoneroMMGenTXFileParseError' in type(e).__name__:
			ymsg(f'\n{type(e).__name__}: {e}')
		return False

class MoneroMMGenFile:

	silent_load = False

	def make_chksum(self,keys=None):
		res = json.dumps(
			dict( (k,v) for k,v in self.data._asdict().items() if (not keys or k in keys) ),
			cls = json_encoder
		)
		return make_chksum_N( res, rounds=1, nchars=self.chksum_nchars, upper=False )

	@property
	def base_chksum(self):
		return self.make_chksum(self.base_chksum_fields)

	@property
	def full_chksum(self):
		return self.make_chksum(self.full_chksum_fields) if self.full_chksum_fields else None

	def check_checksums(self,d_wrap):
		for k in ('base_chksum','full_chksum'):
			a = getattr(self,k)
			if a is not None:
				b = d_wrap[k]
				assert a == b, f'{k} mismatch: {a} != {b}'

	def make_wrapped_data(self,in_data):
		out = {
			'base_chksum': self.base_chksum,
			'full_chksum': self.full_chksum,
			'data': in_data,
		} if self.full_chksum else {
			'base_chksum': self.base_chksum,
			'data': in_data,
		}
		return json.dumps(
			{ self.data_label: out },
			cls = json_encoder,
			indent = 2,
		)

	def extract_data_from_file(self,cfg,fn):
		return json.loads(
			get_data_from_file( cfg, str(fn), self.desc, silent=self.silent_load )
		)[self.data_label]

class MoneroMMGenTX:

	class Base(MoneroMMGenFile):

		data_label = 'MoneroMMGenTX'

		# both base_chksum and full_chksum are used to make the filename stem, so we must not include
		# fields that change when TX is signed and submitted (e.g. ‘sign_time’, ‘submit_time’)
		base_chksum_fields = {
			'op',
			'create_time',
			'network',
			'seed_id',
			'source',
			'dest',
			'amount' }
		full_chksum_fields = {
			'op',
			'create_time',
			'network',
			'seed_id',
			'source',
			'dest',
			'amount',
			'fee',
			'blob' }
		oneline_fs = '{a:7} {b:8} {c:19} {d:13} {e:9} {f:6} {x:2} {g:6} {h:17} {j}'
		oneline_fixed_cols_w = 96 # width of all columns except the last (coin address)
		chksum_nchars = 6
		xmrwallet_tx_data = namedtuple('xmrwallet_tx_data',[
			'op',
			'create_time',
			'sign_time',
			'submit_time',
			'network',
			'seed_id',
			'source',
			'dest',
			'dest_address',
			'txid',
			'amount',
			'priority',
			'fee',
			'blob',
			'metadata',
			'unsigned_txset',
			'signed_txset',
			'complete',
		])

		def __init__(self):
			self.name = type(self).__name__

		@property
		def src_wallet_idx(self):
			return int(self.data.source.split(':')[0])

		def get_info_oneline(self, indent='', addr_w=None):
			d = self.data
			return self.oneline_fs.format(
					a = yellow(d.network),
					b = d.seed_id.hl(),
					c = make_timestr(d.submit_time if d.submit_time is not None else d.create_time),
					d = orange(self.file_id),
					e = purple(d.op.ljust(9)),
					f = red('{}:{}'.format(d.source.wallet,d.source.account).ljust(6)),
					g = red('{}:{}'.format(d.dest.wallet,d.dest.account).ljust(6)) if d.dest else cyan('ext   '),
					h = d.amount.fmt( color=True, iwidth=4, prec=12 ),
					j = d.dest_address.fmt(width=addr_w, color=True) if addr_w else d.dest_address.hl(),
					x = '->'
				)

		def get_info(self, indent='', addr_w=None):
			d = self.data
			pmt_id = d.dest_address.parsed.payment_id
			fs = '\n'.join(list_gen(
				['Info for transaction {a} [Seed ID: {b}. Network: {c}]:'],
				['  TxID:      {d}'],
				['  Created:   {e:19} [{f}]'],
				['  Signed:    {g:19} [{h}]', d.sign_time],
				['  Submitted: {s:19} [{t}]', d.submit_time],
				['  Type:      {i}{S}'],
				['  From:      wallet {j}, account {k}'],
				['  To:        wallet {x}, account {y}, address {z}', d.dest],
				['  Amount:    {m} XMR'],
				['  Priority:  {F}', d.priority],
				['  Fee:       {n} XMR'],
				['  Dest:      {o}'],
				['  Size:      {Z} bytes', d.signed_txset],
				['  Payment ID: {P}', pmt_id],
			))

			from .util2 import format_elapsed_hr
			return fmt(fs,strip_char='\t',indent=indent).format(
					a = orange(self.file_id),
					b = d.seed_id.hl(),
					c = yellow(d.network.upper()),
					d = d.txid.hl(),
					e = make_timestr(d.create_time),
					f = format_elapsed_hr(d.create_time),
					g = make_timestr(d.sign_time) if d.sign_time else None,
					h = format_elapsed_hr(d.sign_time) if d.sign_time else None,
					i = blue(d.op),
					j = d.source.wallet.hl(),
					k = red(f'#{d.source.account}'),
					m = d.amount.hl(),
					F = (Int(d.priority).hl() + f' [{tx_priorities[d.priority]}]') if d.priority else None,
					n = d.fee.hl(),
					o = d.dest_address.hl() if self.cfg.full_address
						else d.dest_address.fmt(width=addr_width, color=True),
					P = pink(pmt_id.hex()) if pmt_id else None,
					s = make_timestr(d.submit_time) if d.submit_time else None,
					S = pink(f" [cold signed{', submitted' if d.complete else ''}]") if d.signed_txset else '',
					t = format_elapsed_hr(d.submit_time) if d.submit_time else None,
					x = d.dest.wallet.hl() if d.dest else None,
					y = red(f'#{d.dest.account}') if d.dest else None,
					z = red(f'#{d.dest.account_address}') if d.dest else None,
					Z = Int(len(d.signed_txset) // 2).hl() if d.signed_txset else None,
				)

		@property
		def file_id(self):
			return (self.base_chksum + ('-' + self.full_chksum if self.full_chksum else '')).upper()

		def write(self,delete_metadata=False,ask_write=True,ask_overwrite=True):
			dict_data = self.data._asdict()
			if delete_metadata:
				dict_data['metadata'] = None

			fn = '{a}-XMR[{b!s}]{c}.{d}'.format(
				a = self.file_id,
				b = self.data.amount,
				c = '' if self.data.network == 'mainnet' else f'.{self.data.network}',
				d = self.ext
			)

			if self.cfg.autosign:
				fn = get_autosign_obj(self.cfg).xmr_tx_dir / fn

			from .fileutil import write_data_to_file
			write_data_to_file(
				cfg                   = self.cfg,
				outfile               = str(fn),
				data                  = self.make_wrapped_data(dict_data),
				desc                  = self.desc,
				ask_write             = ask_write,
				ask_write_default_yes = not ask_write,
				ask_overwrite         = ask_overwrite,
				ignore_opt_outdir     = self.cfg.autosign )

	class New(Base):

		def __init__(self,*args,**kwargs):

			super().__init__()

			assert not args, 'Non-keyword args not permitted'

			if '_in_tx' in kwargs:
				in_data = kwargs.pop('_in_tx').data._asdict()
				in_data.update(kwargs)
			else:
				in_data = kwargs

			d = namedtuple('monero_tx_in_data_tuple',in_data)(**in_data)
			self.cfg = d.cfg

			proto = init_proto( self.cfg, 'xmr', network=d.network, need_amt=True )

			now = int(time.time())

			self.data = self.xmrwallet_tx_data(
				op             = d.op,
				create_time    = now if self.name in ('NewSigned','NewUnsigned') else getattr(d,'create_time',None),
				sign_time      = now if self.name in ('NewSigned','NewColdSigned') else getattr(d,'sign_time',None),
				submit_time    = now if self.name == 'NewSubmitted' else None,
				network        = d.network,
				seed_id        = SeedID(sid=d.seed_id),
				source         = XMRWalletAddrSpec(d.source),
				dest           = None if d.dest is None else XMRWalletAddrSpec(d.dest),
				dest_address   = CoinAddr(proto,d.dest_address),
				txid           = CoinTxID(d.txid),
				amount         = proto.coin_amt(d.amount,from_unit='atomic'),
				priority       = self.cfg.priority if self.name in ('NewSigned','NewUnsigned') else d.priority,
				fee            = proto.coin_amt(d.fee,from_unit='atomic'),
				blob           = d.blob,
				metadata       = d.metadata,
				unsigned_txset = d.unsigned_txset,
				signed_txset   = getattr(d,'signed_txset',None),
				complete       = self.name in ('NewSigned','NewSubmitted'),
			)

	class NewUnsigned(New):
		desc = 'unsigned transaction'
		ext = 'rawtx'
		signed = False

	class NewSigned(New):
		desc = 'signed transaction'
		ext = 'sigtx'
		signed = True

	class NewColdSigned(NewSigned):
		pass

	class NewSubmitted(NewColdSigned):
		desc = 'submitted transaction'
		ext = 'subtx'

	class Completed(Base):
		desc = 'transaction'
		forbidden_fields = ()

		def __init__(self,cfg,fn):

			super().__init__()

			self.cfg = cfg
			self.fn = Path(fn)

			try:
				d_wrap = self.extract_data_from_file( cfg, fn )
			except Exception as e:
				die( 'MoneroMMGenTXFileParseError', f'{type(e).__name__}: {e}\nCould not load transaction file' )

			if 'unsigned_txset' in d_wrap['data']: # post-autosign
				self.full_chksum_fields &= set(d_wrap['data']) # allow for added chksum fields in future
			else:
				self.full_chksum_fields = set(d_wrap['data']) - {'metadata'}

			for key in self.xmrwallet_tx_data._fields: # backwards compat: fill in missing fields
				if not key in d_wrap['data']:
					d_wrap['data'][key] = None

			d = self.xmrwallet_tx_data(**d_wrap['data'])

			if self.name not in ('View','Completed'):
				assert fn.name.endswith('.'+self.ext), 'TX file {fn} has incorrect extension (not {self.ext!r})'
				assert getattr(d,self.req_field), f'{self.name} TX missing required field {self.req_field!r}'
				assert bool(d.sign_time) == self.signed, '{a} has {b}sign time!'.format(
					a = self.desc,
					b = 'no ' if self.signed else'' )
				for f in self.forbidden_fields:
					assert not getattr(d,f), f'{self.name} TX mismatch: contains forbidden field {f!r}'

			proto = init_proto( cfg, 'xmr', network=d.network, need_amt=True )

			self.data = self.xmrwallet_tx_data(
				op             = d.op,
				create_time    = d.create_time,
				sign_time      = d.sign_time,
				submit_time    = d.submit_time,
				network        = d.network,
				seed_id        = SeedID(sid=d.seed_id),
				source         = XMRWalletAddrSpec(d.source),
				dest           = None if d.dest is None else XMRWalletAddrSpec(d.dest),
				dest_address   = CoinAddr(proto,d.dest_address),
				txid           = CoinTxID(d.txid),
				amount         = proto.coin_amt(d.amount),
				priority       = d.priority,
				fee            = proto.coin_amt(d.fee),
				blob           = d.blob,
				metadata       = d.metadata,
				unsigned_txset = d.unsigned_txset,
				signed_txset   = d.signed_txset,
				complete       = d.complete,
			)

			self.check_checksums(d_wrap)

	class Unsigned(Completed):
		desc = 'unsigned transaction'
		ext = 'rawtx'
		signed = False
		req_field = 'unsigned_txset'
		forbidden_fields = ('signed_txset',)

	class Signed(Completed):
		desc = 'signed transaction'
		ext = 'sigtx'
		signed = True
		req_field = 'blob'
		forbidden_fields = ('signed_txset','unsigned_txset')

	class ColdSigned(Signed):
		req_field = 'signed_txset'
		forbidden_fields = ()

	class Submitted(ColdSigned):
		desc = 'submitted transaction'
		ext = 'subtx'
		silent_load = True

	class View(Completed):
		silent_load = True

class MoneroWalletOutputsFile:

	class Base(MoneroMMGenFile):

		desc = 'wallet outputs'
		data_label = 'MoneroMMGenWalletOutputsFile'
		base_chksum_fields = {'seed_id','wallet_index','outputs_data_hex',}
		full_chksum_fields = {'seed_id','wallet_index','outputs_data_hex','signed_key_images'}
		fn_fs = '{a}-outputs-{b}.{c}'
		ext_offset = 25 # len('-outputs-') + len(chksum) ({b})
		chksum_nchars = 16
		data_tuple = namedtuple('wallet_outputs_data',[
			'seed_id',
			'wallet_index',
			'outputs_data_hex',
			'signed_key_images',
			'sign',
			'imported',
		])

		def __init__(self,cfg):
			self.name = type(self).__name__
			self.cfg = cfg

		def write(self, add_suf='', quiet=False):
			from .fileutil import write_data_to_file
			write_data_to_file(
				cfg               = self.cfg,
				outfile           = str(self.get_outfile( self.cfg, self.wallet_fn )) + add_suf,
				data              = self.make_wrapped_data(self.data._asdict()),
				desc              = self.desc,
				ask_overwrite     = False,
				quiet             = quiet,
				ignore_opt_outdir = True)

		def get_outfile(self,cfg,wallet_fn):
			return (
				get_autosign_obj(cfg).xmr_outputs_dir if cfg.autosign else
				wallet_fn.parent ) / self.fn_fs.format(
					a = wallet_fn.name,
					b = self.base_chksum,
					c = self.ext,
				)

		def get_wallet_fn(self,fn):
			assert fn.name.endswith(f'.{self.ext}'), (
				f'{self.name}: filename does not end with {"."+self.ext!r}'
			)
			return fn.parent / fn.name[:-(len(self.ext)+self.ext_offset+1)]

		def get_info(self,indent=''):
			if self.data.signed_key_images is not None:
				data = self.data.signed_key_images or []
				return f'{indent}{self.wallet_fn.name}: {len(data)} signed key image{suf(data)}'
			else:
				return f'{indent}{self.wallet_fn.name}: no key images'

	class New(Base):
		ext = 'raw'

		def __init__(self, parent, wallet_fn, data, wallet_idx=None, sign=False):
			super().__init__(parent.cfg)
			self.wallet_fn = wallet_fn
			init_data = dict.fromkeys(self.data_tuple._fields)
			init_data.update({
				'seed_id':      parent.kal.al_id.sid,
				'wallet_index': wallet_idx or parent.get_idx_from_fn(wallet_fn),
			})
			if sign:
				init_data['sign'] = sign
			init_data.update({k:v for k,v in data.items() if k in init_data})
			self.data = self.data_tuple(**init_data)

	class Completed(New):

		def __init__( self, parent, fn=None, wallet_fn=None ):
			def check_equal(desc,a,b):
				assert a == b, f'{desc} mismatch: {a} (from file) != {b} (from filename)'
			fn = fn or self.get_outfile( parent.cfg, wallet_fn )
			wallet_fn = wallet_fn or self.get_wallet_fn(fn)
			d_wrap = self.extract_data_from_file( parent.cfg, fn )
			data = d_wrap['data']
			check_equal( 'Seed ID', data['seed_id'], parent.kal.al_id.sid )
			wallet_idx = parent.get_idx_from_fn(wallet_fn)
			check_equal( 'Wallet index', data['wallet_index'], wallet_idx )
			super().__init__(
				parent     = parent,
				wallet_fn  = wallet_fn,
				data       = data,
				wallet_idx = wallet_idx,
			)
			self.check_checksums(d_wrap)

		@classmethod
		def find_fn_from_wallet_fn(cls,cfg,wallet_fn,ret_on_no_match=False):
			path = get_autosign_obj(cfg).xmr_outputs_dir or Path()
			pat = cls.fn_fs.format(
				a = wallet_fn.name,
				b = f'[0-9a-f]{{{cls.chksum_nchars}}}\\',
				c = cls.ext,
			)
			matches = [f for f in path.iterdir() if re.match(pat,f.name)]
			if not matches and ret_on_no_match:
				return None
			if not matches or len(matches) > 1:
				die(2,"{a} matching pattern {b!r} found in '{c}'!".format(
					a = 'No files' if not matches else 'More than one file',
					b = pat,
					c = path
				))
			return matches[0]

	class Unsigned(Completed):
		pass

	class SignedNew(New):
		desc = 'signed key images'
		ext = 'sig'

	class Signed(Completed,SignedNew):
		pass

class MoneroWalletDumpFile:

	class Base:
		desc = 'Monero wallet dump'
		data_label = 'MoneroMMGenWalletDumpFile'
		base_chksum_fields = {'seed_id','wallet_index','wallet_metadata'}
		full_chksum_fields = None
		ext = 'dump'
		ext_offset = 0
		data_tuple = namedtuple('wallet_dump_data',[
			'seed_id',
			'wallet_index',
			'wallet_metadata',
		])
		def get_outfile(self,cfg,wallet_fn):
			return wallet_fn.parent / f'{wallet_fn.name}.{self.ext}'

	class New(Base,MoneroWalletOutputsFile.New):
		pass

	class Completed(Base,MoneroWalletOutputsFile.Completed):
		pass

class MoneroWalletOps:

	ops = (
		'create',
		'create_offline',
		'sync',
		'list',
		'view',
		'listview',
		'new',
		'transfer',
		'sweep',
		'sweep_all',
		'relay',
		'txview',
		'txlist',
		'label',
		'sign',
		'submit',
		'resubmit',
		'abort',
		'dump',
		'restore',
		'export_outputs',
		'import_key_images' )

	kafile_arg_ops = (
		'create',
		'sync',
		'list',
		'view',
		'listview',
		'label',
		'new',
		'transfer',
		'sweep',
		'sweep_all',
		'dump',
		'restore' )

	opts = (
		'wallet_dir',
		'daemon',
		'tx_relay_daemon',
		'use_internal_keccak_module',
		'hash_preset',
		'restore_height',
		'no_start_wallet_daemon',
		'no_stop_wallet_daemon',
		'no_relay',
		'watch_only',
		'autosign',
		'skip_empty_accounts',
		'skip_empty_addresses')

	pat_opts = ('daemon','tx_relay_daemon')

	class base(MMGenObject):

		opts = ('wallet_dir',)
		trust_monerod = False
		do_umount = True

		def __init__(self,cfg,uarg_tuple):

			def gen_classes():
				for cls in type(self).__mro__:
					yield cls
					if cls.__name__ == 'base':
						break

			self.name = type(self).__name__
			self.cfg = cfg
			classes = tuple(gen_classes())
			self.opts = tuple(set(opt for cls in classes for opt in cls.opts))

			if not hasattr(self,'stem'):
				self.stem = self.name

			global uarg, uarg_info, fmt_amt, hl_amt, addr_width

			uarg = uarg_tuple
			uarg_info = xmrwallet_uarg_info

			def fmt_amt(amt):
				return self.proto.coin_amt(amt,from_unit='atomic').fmt( iwidth=5, prec=12, color=True )
			def hl_amt(amt):
				return self.proto.coin_amt(amt,from_unit='atomic').hl()

			addr_width = 95 if self.cfg.full_address else 17

			self.proto = init_proto(cfg, 'xmr', network=self.cfg.network, need_amt=True)

			id_cur = None
			for cls in classes:
				if id(cls.check_uopts) != id_cur:
					cls.check_uopts(self)
					id_cur = id(cls.check_uopts)

			id_cur = None
			for cls in classes:
				if id(cls.pre_init_action) != id_cur:
					cls.pre_init_action(self)
					id_cur = id(cls.pre_init_action)

			if cfg.autosign:
				self.asi = get_autosign_obj(cfg)

		def check_uopts(self):

			def check_pat_opt(name):
				val = getattr(self.cfg,name)
				if not re.fullmatch( uarg_info[name].pat, val, re.ASCII ):
					die(1,'{!r}: invalid value for --{}: it must have format {!r}'.format(
						val,
						name.replace('_','-'),
						uarg_info[name].annot
					))

			for attr in self.cfg.__dict__:
				if attr in MoneroWalletOps.opts and not attr in self.opts:
					die(1,'Option --{} not supported for {!r} operation'.format(
						attr.replace('_','-'),
						self.name,
					))

			for opt in MoneroWalletOps.pat_opts:
				if getattr(self.cfg,opt,None):
					check_pat_opt(opt)

		def parse_tx_relay_opt(self):
			return re.fullmatch(
				uarg_info['tx_relay_daemon'].pat,
				self.cfg.tx_relay_daemon,
				re.ASCII )

		def display_tx_relay_info(self,indent=''):
			m = self.parse_tx_relay_opt()
			msg(fmt(f"""
				TX relay info:
				  Host:  {blue(m[1])}
				  Proxy: {blue(m[2] or 'None')}
				""",strip_char='\t',indent=indent))

		def pre_init_action(self):
			pass

		def post_main_success(self):
			pass

		def post_main_failure(self):
			pass

		async def stop_wallet_daemon(self):
			pass

		def post_mount_action(self):
			pass

		def mount_removable_device(self):
			if self.cfg.autosign:
				if not self.asi.get_insert_status():
					die(1,'Removable device not present!')
				if self.do_umount:
					import atexit
					atexit.register(lambda: self.asi.do_umount())
				self.asi.do_mount()
				self.post_mount_action()

	class wallet(base):

		opts = (
			'use_internal_keccak_module',
			'hash_preset',
			'daemon',
			'no_start_wallet_daemon',
			'no_stop_wallet_daemon',
			'autosign',
			'watch_only',
		)
		wallet_offline = False
		wallet_exists = True
		start_daemon = True
		skip_wallet_check = False # for debugging

		def __init__(self,cfg,uarg_tuple):

			def wallet_exists(fn):
				try:
					fn.stat()
				except:
					return False
				else:
					return True

			def check_wallets():
				for d in self.addr_data:
					fn = self.get_wallet_fn(d)
					exists = wallet_exists(fn)
					if exists and not self.wallet_exists:
						die(1, f'Wallet ‘{fn}’ already exists!')
					elif not exists and self.wallet_exists:
						die(1, f'Wallet ‘{fn}’ not found!')

			super().__init__(cfg,uarg_tuple)

			if self.cfg.offline or (self.name == 'create' and self.cfg.restore_height is None):
				self.wallet_offline = True

			self.wd = MoneroWalletDaemon(
				cfg         = self.cfg,
				proto       = self.proto,
				wallet_dir  = self.cfg.wallet_dir or '.',
				test_suite  = self.cfg.test_suite,
				monerod_addr = self.cfg.daemon or None,
				trust_monerod = self.trust_monerod,
				test_monerod = not self.wallet_offline,
			)

			if self.wallet_offline:
				self.wd.usr_daemon_args = ['--offline']

			self.c = MoneroWalletRPCClient(
				cfg             = self.cfg,
				daemon          = self.wd,
				test_connection = False,
			)

			if self.cfg.offline:
				from .wallet import Wallet
				self.seed_src = Wallet(
					cfg           = cfg,
					fn            = uarg.infile,
					ignore_in_fmt = True )

				gmsg('\nCreating ephemeral key-address list for offline wallets')
				self.kal = KeyAddrList(
					cfg       = cfg,
					proto     = self.proto,
					seed      = self.seed_src.seed,
					addr_idxs = uarg.wallets,
					skip_chksum_msg = True )
			else:
				self.mount_removable_device()
				# with watch_only, make a second attempt to open the file as KeyAddrList:
				for first_try in (True,False):
					try:
						self.kal = (ViewKeyAddrList if (self.cfg.watch_only and first_try) else KeyAddrList)(
							cfg      = cfg,
							proto    = self.proto,
							addrfile = str(self.autosign_viewkey_addr_file) if self.cfg.autosign else uarg.infile,
							key_address_validity_check = True,
							skip_chksum_msg = True )
						break
					except:
						if first_try:
							msg(f"Attempting to open '{uarg.infile}' as key-address list")
							continue
						raise

			self.create_addr_data()

			if not self.skip_wallet_check:
				check_wallets()

			if self.start_daemon and not self.cfg.no_start_wallet_daemon:
				async_run(self.restart_wallet_daemon())

		@classmethod
		def get_idx_from_fn(cls,fn):
			return int( re.match(r'[0-9a-fA-F]{8}-(\d+)-Monero(WatchOnly)?Wallet.*',fn.name)[1] )

		def pre_init_action(self):
			if self.cfg.skip_empty_accounts:
				msg(orange('Skipping display of empty accounts where applicable'))
			if self.cfg.skip_empty_addresses:
				msg(orange('Skipping display of empty used addresses where applicable'))

		def get_coin_daemon_rpc(self):

			host,port = self.cfg.daemon.split(':') if self.cfg.daemon else ('localhost',self.wd.monerod_port)

			from .daemon import CoinDaemon
			return MoneroRPCClient(
				cfg    = self.cfg,
				proto  = self.proto,
				daemon = CoinDaemon( self.cfg, 'xmr' ),
				host   = host,
				port   = int(port),
				user   = None,
				passwd = None )

		@property
		def autosign_viewkey_addr_file(self):
			from .addrfile import ViewKeyAddrFile
			flist = [f for f in self.asi.xmr_dir.iterdir() if f.name.endswith(ViewKeyAddrFile.ext)]
			if len(flist) != 1:
				die(2,
					"{a} viewkey-address files found in autosign mountpoint directory '{b}'!\n".format(
						a = 'Multiple' if flist else 'No',
						b = self.asi.xmr_dir
					)
					+ 'Have you run ‘mmgen-autosign setup’ on your offline machine with the --xmrwallets option?'
				)
			else:
				return flist[0]

		def create_addr_data(self):
			if uarg.wallets:
				idxs = AddrIdxList(uarg.wallets)
				self.addr_data = [d for d in self.kal.data if d.idx in idxs]
				if len(self.addr_data) != len(idxs):
					die(1,f'List {uarg.wallets!r} contains addresses not present in supplied key-address file')
			else:
				self.addr_data = self.kal.data

		async def restart_wallet_daemon(self):
			atexit.register(lambda: async_run(self.stop_wallet_daemon()))
			await self.c.restart_daemon()

		async def stop_wallet_daemon(self):
			if not self.cfg.no_stop_wallet_daemon:
				try:
					await self.c.stop_daemon()
				except KeyboardInterrupt:
					ymsg('\nForce killing wallet daemon')
					self.c.daemon.force_kill = True
					self.c.daemon.stop()

		def get_wallet_fn(self,data,watch_only=None):
			if watch_only is None:
				watch_only = self.cfg.watch_only
			return Path(
				(self.cfg.wallet_dir or '.'),
				'{a}-{b}-Monero{c}Wallet{d}'.format(
					a = self.kal.al_id.sid,
					b = data.idx,
					c = 'WatchOnly' if watch_only else '',
					d = f'.{self.cfg.network}' if self.cfg.network != 'mainnet' else '')
			)

		@property
		def add_wallet_desc(self):
			return 'offline signing ' if self.cfg.offline else 'watch-only ' if self.cfg.watch_only else ''

		async def main(self):
			gmsg('\n{a}ing {b} {c}wallet{d}'.format(
				a = self.stem.capitalize(),
				b = len(self.addr_data),
				c = self.add_wallet_desc,
				d = suf(self.addr_data) ))
			processed = 0
			for n,d in enumerate(self.addr_data): # [d.sec,d.addr,d.wallet_passwd,d.viewkey]
				fn = self.get_wallet_fn(d)
				gmsg('\n{a}ing wallet {b}/{c} ({d})'.format(
					a = self.stem.capitalize(),
					b = n + 1,
					c = len(self.addr_data),
					d = fn.name,
				))
				processed += await self.process_wallet(
					d,
					fn,
					last = n==len(self.addr_data)-1 )
			gmsg(f'\n{processed} wallet{suf(processed)} {self.stem}ed\n')
			return processed

		def head_msg(self,wallet_idx,fn):
			gmsg('\n{a} {b}wallet #{c} ({d})'.format(
				a = self.action.capitalize(),
				b = self.add_wallet_desc,
				c = wallet_idx,
				d = fn.name
			))

		class rpc:

			def __init__(self,parent,d):
				self.parent = parent
				self.cfg = parent.cfg
				self.proto = parent.proto
				self.c = parent.c
				self.d = d
				self.fn = parent.get_wallet_fn(d)
				self.new_tx_cls = (
					MoneroMMGenTX.NewUnsigned if self.cfg.watch_only else
					MoneroMMGenTX.NewSigned )

			def open_wallet(self,desc=None,refresh=True):
				add_desc = desc + ' ' if desc else self.parent.add_wallet_desc
				gmsg_r(f'\n  Opening {add_desc}wallet...')
				self.c.call( # returns {}
					'open_wallet',
					filename = self.fn.name,
					password = self.d.wallet_passwd )
				gmsg('done')

				if refresh:
					gmsg_r(f'  Refreshing {add_desc}wallet...')
					ret = self.c.call('refresh')
					gmsg('done')
					if ret['received_money']:
						msg('  Wallet has received funds')

			def close_wallet(self,desc):
				gmsg_r(f'\n  Closing {desc} wallet...')
				self.c.call('close_wallet')
				gmsg_r('done')

			async def stop_wallet(self,desc):
				msg(f'Stopping {self.c.daemon.desc} on port {self.c.daemon.bind_port}')
				gmsg_r(f'\n  Stopping {desc} wallet...')
				await self.c.stop_daemon(quiet=True) # closes wallet
				gmsg_r('done')

			def gen_accts_info(self, accts_data, addrs_data, indent='    ', skip_empty_ok=False):
				fs = indent + '  {I:<3} {A} {N} {B} {L}'
				yield indent + f'Accounts of wallet {self.fn.name}:'
				yield fs.format(
					I = '',
					A = 'Base Address'.ljust(addr_width),
					N = 'nAddrs',
					B = '  Unlocked Balance',
					L = 'Label')
				for i,e in enumerate(accts_data['subaddress_accounts']):
					if skip_empty_ok and self.cfg.skip_empty_accounts and not e['unlocked_balance']:
						continue
					ca = CoinAddr(self.proto, e['base_address'])
					yield fs.format(
						I = str(e['account_index']),
						A = ca.hl() if self.cfg.full_address else ca.fmt(color=True, width=addr_width),
						N = red(str(len(addrs_data[i]['addresses'])).ljust(6)),
						B = fmt_amt(e['unlocked_balance']),
						L = pink(e['label']))

			def get_wallet_data(self, print=True, skip_empty_ok=False):
				accts_data = self.c.call('get_accounts')
				addrs_data = [
					self.c.call('get_address',account_index=i)
						for i in range(len(accts_data['subaddress_accounts']))
				]
				if print:
					msg('\n' + '\n'.join(self.gen_accts_info(accts_data, addrs_data, skip_empty_ok=skip_empty_ok)))
				bals_data = self.c.call('get_balance', all_accounts=True)
				return namedtuple('wallet_data', ['accts_data', 'addrs_data', 'bals_data'])(
					accts_data, addrs_data, bals_data)

			def create_acct(self,label=None):
				msg('\n    Creating new account...')
				ret = self.c.call('create_account', label=label)
				msg('      Index:   {}'.format( pink(str(ret['account_index'])) ))
				msg('      Address: {}'.format( cyan(ret['address']) ))
				return (ret['account_index'], ret['address'])

			def get_last_acct(self,accts_data):
				msg('\n    Getting last account...')
				ret = accts_data['subaddress_accounts'][-1]
				msg('      Index:   {}'.format( pink(str(ret['account_index'])) ))
				msg('      Address: {}'.format( cyan(ret['base_address']) ))
				return (ret['account_index'], ret['base_address'])

			def print_acct_addrs(self, wallet_data, account):
				msg('\n      Addresses of account #{} ({}):'.format(
					account,
					wallet_data.accts_data['subaddress_accounts'][account]['label']))
				msg('\n'.join(gen_acct_addr_info(self, wallet_data, account, indent='        ')))
				return wallet_data.addrs_data[account]['addresses']

			def create_new_addr(self, account, label):
				msg_r('\n    Creating new address: ')
				ret = self.c.call('create_address', account_index=account, label=label or '')
				msg(cyan(ret['address']))
				return ret['address']

			def get_last_addr(self, account, wallet_data, display=True):
				if display:
					msg('\n    Getting last address:')
				acct_addrs = wallet_data.addrs_data[account]['addresses']
				addr = acct_addrs[-1]['address']
				if display:
					msg('      ' + cyan(addr))
				return (addr, len(acct_addrs) - 1)

			def set_label(self,account,address_idx,label):
				return self.c.call(
					'label_address',
					index = { 'major': account, 'minor': address_idx },
					label = label
				)

			def make_transfer_tx(self,account,addr,amt):
				res = self.c.call(
					'transfer',
					account_index = account,
					destinations = [{
						'amount':  amt.to_unit('atomic'),
						'address': addr
					}],
					priority = self.cfg.priority or None,
					do_not_relay = True,
					get_tx_hex = True,
					get_tx_metadata = True
				)
				return self.new_tx_cls(
					cfg            = self.cfg,
					op             = self.parent.name,
					network        = self.proto.network,
					seed_id        = self.parent.kal.al_id.sid,
					source         = XMRWalletAddrSpec(self.parent.source.idx,self.parent.account,None),
					dest           = None,
					dest_address   = addr,
					txid           = res['tx_hash'],
					amount         = res['amount'],
					fee            = res['fee'],
					blob           = res['tx_blob'],
					metadata       = res['tx_metadata'],
					unsigned_txset = res['unsigned_txset'] if self.cfg.watch_only else None,
				)

			def make_sweep_tx(self, account, dest_acct, dest_addr_idx, addr, addrs_data):
				res = self.c.call(
					'sweep_all',
					address = addr,
					account_index = account,
					subaddr_indices = list(range(len(addrs_data[account]['addresses'])))
						if self.parent.name == 'sweep_all' else [],
					priority = self.cfg.priority or None,
					do_not_relay = True,
					get_tx_hex = True,
					get_tx_metadata = True
				)

				if len(res['tx_hash_list']) > 1:
					die(3,'More than one TX required.  Cannot perform this sweep')

				return self.new_tx_cls(
					cfg            = self.cfg,
					op             = self.parent.name,
					network        = self.proto.network,
					seed_id        = self.parent.kal.al_id.sid,
					source         = XMRWalletAddrSpec(self.parent.source.idx,self.parent.account,None),
					dest           = XMRWalletAddrSpec(
										(self.parent.dest or self.parent.source).idx,
										dest_acct,
										dest_addr_idx),
					dest_address   = addr,
					txid           = res['tx_hash_list'][0],
					amount         = res['amount_list'][0],
					fee            = res['fee_list'][0],
					blob           = res['tx_blob_list'][0],
					metadata       = res['tx_metadata_list'][0],
					unsigned_txset = res['unsigned_txset'] if self.cfg.watch_only else None,
				)

			def relay_tx(self,tx_hex):
				ret = self.c.call('relay_tx',hex=tx_hex)
				try:
					msg('\n    Relayed {}'.format( CoinTxID(ret['tx_hash']).hl() ))
				except:
					msg(f'\n   Server returned: {ret!s}')

	class create(wallet):
		stem    = 'creat'
		wallet_exists = False
		opts    = ('restore_height',)

		def check_uopts(self):
			if self.cfg.restore_height != 'current':
				if int(self.cfg.restore_height or 0) < 0:
					die(1,f'{self.cfg.restore_height}: invalid value for --restore-height (less than zero)')

		async def process_wallet(self,d,fn,last):
			msg_r('') # for pexpect

			if self.cfg.restore_height == 'current':
				restore_height = self.get_coin_daemon_rpc().call_raw('get_height')['height']
			else:
				restore_height = self.cfg.restore_height

			if self.cfg.watch_only:
				ret = self.c.call(
					'generate_from_keys',
					filename       = fn.name,
					password       = d.wallet_passwd,
					address        = d.addr,
					viewkey        = d.viewkey,
					restore_height = restore_height )
			else:
				from .xmrseed import xmrseed
				ret = self.c.call(
					'restore_deterministic_wallet',
					filename       = fn.name,
					password       = d.wallet_passwd,
					seed           = xmrseed().fromhex(d.sec.wif,tostr=True),
					restore_height = restore_height,
					language       = 'English' )

			pp_msg(ret) if self.cfg.debug else msg(f'  Address: {ret["address"]}')
			return True

	class create_offline(create):

		def __init__(self,cfg,uarg_tuple):

			super().__init__(cfg,uarg_tuple)

			gmsg('\nCreating viewkey-address file for watch-only wallets')
			vkal = ViewKeyAddrList(
				cfg       = self.cfg,
				proto     = self.proto,
				addrfile  = None,
				addr_idxs = uarg.wallets,
				seed      = self.seed_src.seed,
				skip_chksum_msg = True )
			vkf = vkal.file

			# before writing viewkey-address file, shred any old ones in the directory:
			for f in Path(self.asi.xmr_dir).iterdir():
				if f.name.endswith(vkf.ext):
					from .fileutil import shred_file
					msg(f"\nShredding old viewkey-address file '{f}'")
					shred_file( f, verbose=self.cfg.verbose )

			vkf.write(outdir=self.asi.xmr_dir)

	class restore(create):
		wallet_offline = True

		def check_uopts(self):
			if self.cfg.restore_height is not None:
				die(1,'--restore-height must be unset when running the ‘restore’ command')

		async def process_wallet(self,d,fn,last):

			def get_dump_data():
				def gen():
					for fn in [self.get_wallet_fn(d,watch_only=wo) for wo in (True,False)]:
						ret = fn.parent / (fn.name + '.dump')
						if ret.exists():
							yield ret
				dump_fns = tuple(gen())
				if not dump_fns:
					die(1,f"No suitable dump file found for '{fn}'")
				elif len(dump_fns) > 1:
					ymsg(f"Warning: more than one dump file found for '{fn}' - using the first!")
				return MoneroWalletDumpFile.Completed(
					parent = self,
					fn     = dump_fns[0] ).data._asdict()['wallet_metadata']

			def restore_accounts():
				bmsg('  Restoring accounts:')
				for acct_idx,acct_data in enumerate(data[1:],1):
					msg(fs.format(acct_idx, 0, acct_data['address']))
					self.c.call('create_account')

			def restore_subaddresses():
				bmsg('  Restoring subaddresses:')
				for acct_idx,acct_data in enumerate(data):
					for addr_idx,addr_data in enumerate(acct_data['addresses'][1:],1):
						msg(fs.format(acct_idx, addr_idx, addr_data['address']))
						self.c.call( 'create_address', account_index=acct_idx )

			def restore_labels():
				bmsg('  Restoring labels:')
				for acct_idx,acct_data in enumerate(data):
					for addr_idx,addr_data in enumerate(acct_data['addresses']):
						addr_data['used'] = False # do this so that restored data matches
						msg(fs.format(acct_idx, addr_idx, addr_data['label']))
						self.c.call(
							'label_address',
							index = { 'major': acct_idx, 'minor': addr_idx },
							label = addr_data['label'],
						)

			def make_format_str():
				return '    acct {:O>%s}, addr {:O>%s} [{}]' % (
					len(str( len(data) - 1 )),
					len(str( max(len(acct_data['addresses']) for acct_data in data) - 1))
				)

			def check_restored_data():
				restored_data = h.get_wallet_data(print=False).addrs_data
				if restored_data != data:
					rmsg('Restored data does not match original dump!  Dumping bad data.')
					MoneroWalletDumpFile.New(
						parent    = self,
						wallet_fn = fn,
						data      = {'wallet_metadata': restored_data}
					).write(add_suf='.bad')
					die(3,'Fatal error')

			await super().process_wallet(d,fn,last)

			h = self.rpc(self,d)
			h.open_wallet('newly created')

			msg('')
			data = get_dump_data()
			fs = make_format_str()

			gmsg('\nRestoring accounts, subaddresses and labels from dump file:\n')

			restore_accounts()
			restore_subaddresses()
			restore_labels()

			check_restored_data()

			return True

	class sync(wallet):
		opts = ('rescan_blockchain', 'skip_empty_accounts', 'skip_empty_addresses')

		def check_uopts(self):
			if self.cfg.rescan_blockchain and self.cfg.watch_only:
				die(1,f'Operation {self.name!r} does not support --rescan-blockchain with watch-only wallets')

		def __init__(self,cfg,uarg_tuple):

			super().__init__(cfg,uarg_tuple)

			if not self.wallet_offline:
				self.dc = self.get_coin_daemon_rpc()

			self.wallets_data = {}

		async def process_wallet(self,d,fn,last):

			chain_height = self.dc.call_raw('get_height')['height']
			msg(f'  Chain height: {chain_height}')

			t_start = time.time()

			msg_r('  Opening wallet...')
			self.c.call(
				'open_wallet',
				filename = fn.name,
				password = d.wallet_passwd )
			msg('done')

			msg_r('  Getting wallet height (be patient, this could take a long time)...')
			wallet_height = self.c.call('get_height')['height']
			msg_r('\r' + ' '*68 + '\r')
			msg(f'  Wallet height: {wallet_height}        ')

			behind = chain_height - wallet_height
			if behind > 1000:
				msg_r(f'  Wallet is {behind} blocks behind chain tip.  Please be patient.  Syncing...')

			ret = self.c.call('refresh')

			if behind > 1000:
				msg('done')

			if ret['received_money']:
				msg('  Wallet has received funds')

			for i in range(2):
				wallet_height = self.c.call('get_height')['height']
				if wallet_height >= chain_height:
					break
				ymsg(f'  Wallet failed to sync (wallet height [{wallet_height}] < chain height [{chain_height}])')
				if i or not self.cfg.rescan_blockchain:
					break
				msg_r('  Rescanning blockchain, please be patient...')
				self.c.call('rescan_blockchain')
				self.c.call('refresh')
				msg('done')

			t_elapsed = int(time.time() - t_start)

			wd = self.rpc(self, d).get_wallet_data(print=False, skip_empty_ok=True)

			msg('  Balance: {} Unlocked balance: {}'.format(
				hl_amt(wd.accts_data['total_balance']),
				hl_amt(wd.accts_data['total_unlocked_balance']),
			))

			self.wallets_data[fn.name] = wd

			msg(f'  Wallet height: {wallet_height}')
			msg(f'  Sync time: {t_elapsed//60:02}:{t_elapsed%60:02}')

			if not last:
				self.c.call('close_wallet')

			return wallet_height >= chain_height

		def gen_body(self, wallets_data):
			for wnum, (_, wallet_data) in enumerate(wallets_data.items()):
				yield from self.rpc(self, self.addr_data[wnum]).gen_accts_info(
					wallet_data.accts_data,
					wallet_data.addrs_data,
					indent = '',
					skip_empty_ok = True)
				yield ''

		def post_main_success(self):

			def gen_info(data):
				yield from self.gen_body(data)

				col1_w = max(map(len, data)) + 1
				fs = '{:%s} {} {}' % col1_w
				tbals = [0, 0]
				yield fs.format('Wallet', 'Balance           ', 'Unlocked Balance')

				for k in data:
					b  = data[k].accts_data['total_balance']
					ub = data[k].accts_data['total_unlocked_balance']
					yield fs.format(k + ':', fmt_amt(b), fmt_amt(ub))
					tbals[0] += b
					tbals[1] += ub

				yield fs.format('-'*col1_w, '-'*18, '-'*18)
				yield fs.format('TOTAL:', fmt_amt(tbals[0]), fmt_amt(tbals[1]))

			self.cfg._util.stdout_or_pager('\n'.join(gen_info(self.wallets_data)) + '\n')

	class list(sync):
		stem = 'sync'

		def gen_body(self, wallets_data):
			for (wallet_fn, wallet_data) in wallets_data.items():
					ad = wallet_data.accts_data['subaddress_accounts']
					yield green(f'Wallet {wallet_fn}:')
					for account in range(len(wallet_data.addrs_data)):
						bal = ad[account]['unlocked_balance']
						if self.cfg.skip_empty_accounts and not bal:
							continue
						yield ''
						yield '  Account #{a} [{b} {c}]'.format(
							a = account,
							b = self.proto.coin_amt(bal, from_unit='atomic').hl(),
							c = self.proto.coin_amt.hlc('XMR'))
						yield from gen_acct_addr_info(self, wallet_data, account, indent='  ')

					yield ''

	class view(sync):
		stem = 'open'
		opts = ()
		wallet_offline = True

		def pre_init_action(self):
			ymsg('Running in offline mode. Balances may be out of date!')

		async def process_wallet(self,d,fn,last):

			self.c.call(
				'open_wallet',
				filename = fn.name,
				password = d.wallet_passwd)

			wallet_height = self.c.call('get_height')['height']
			msg(f'  Wallet height: {wallet_height}')

			self.wallets_data[fn.name] = self.rpc(self, d).get_wallet_data(print=False, skip_empty_ok=True)

			if not last:
				self.c.call('close_wallet')

			return True

	class listview(view, list):
		pass

	class spec(wallet): # virtual class

		def create_addr_data(self):
			m = re.fullmatch(uarg_info[self.spec_id].pat,uarg.spec,re.ASCII)
			if not m:
				fs = "{!r}: invalid {!r} arg: for {} operation, it must have format {!r}"
				die(1,fs.format( uarg.spec, self.spec_id, self.name, uarg_info[self.spec_id].annot ))

			def gen():
				for i,k in self.spec_key:
					if m[i] is None:
						setattr(self,k,None)
					else:
						idx = int(m[i])
						try:
							res = self.kal.entry(idx)
						except:
							die(1,f'Supplied key-address file does not contain address {self.kal.al_id.sid}:{idx}')
						else:
							setattr(self,k,res)
							yield res

			self.addr_data = list(gen())
			self.account = None if m[2] is None else int(m[2])

			def strip_quotes(s):
				if s and s[0] in ("'",'"'):
					if s[-1] != s[0] or len(s) < 2:
						die(1,f'{s!r}: unbalanced quotes in label string!')
					return s[1:-1]
				else:
					return s # None or empty string

			if self.name in ('sweep', 'sweep_all'):
				self.dest_acct = None if m[4] is None else int(m[4])
			elif self.name == 'transfer':
				self.dest_addr = CoinAddr(self.proto,m[3])
				self.amount = self.proto.coin_amt(m[4])
			elif self.name == 'new':
				self.label = strip_quotes(m[3])
			elif self.name == 'label':
				self.address_idx = int(m[3])
				self.label = strip_quotes(m[4])

	class sweep(spec):
		spec_id  = 'sweep_spec'
		spec_key = ( (1,'source'), (3,'dest') )
		opts = (
			'no_relay',
			'tx_relay_daemon',
			'watch_only',
			'priority',
			'skip_empty_accounts',
			'skip_empty_addresses')
		sweep_type = 'single-address'

		def check_uopts(self):
			if self.cfg.tx_relay_daemon and (self.cfg.no_relay or self.cfg.autosign):
				die(1,'--tx-relay-daemon makes no sense in this context!')

			if self.cfg.priority and self.cfg.priority not in list(tx_priorities):
				die(1, '{}: invalid parameter for --priority (valid params: {})'.format(
					self.cfg.priority,
					fmt_dict(tx_priorities, fmt='square_compact')))

		def init_tx_relay_daemon(self):

			m = self.parse_tx_relay_opt()

			wd2 = MoneroWalletDaemon(
				cfg         = self.cfg,
				proto       = self.proto,
				wallet_dir  = self.cfg.wallet_dir or '.',
				test_suite  = self.cfg.test_suite,
				monerod_addr = m[1],
				proxy       = m[2] )

			if self.cfg.test_suite:
				wd2.usr_daemon_args = ['--daemon-ssl-allow-any-cert']

			wd2.start()

			self.c = MoneroWalletRPCClient(
				cfg    = self.cfg,
				daemon = wd2 )

		def create_tx(self, h, wallet_data):

			def create_new_addr_maybe(h, account, label):
				if keypress_confirm(self.cfg, f'\nCreate new address for account #{account}?'):
					return h.create_new_addr(account, label)
				elif not keypress_confirm(self.cfg, f'Sweep to last existing address of account #{account}?'):
					die(1,'Exiting at user request')
				return None

			dest_addr_chk = None

			if self.dest is None: # sweep to same account
				dest_acct = self.account
				dest_addr_chk = create_new_addr_maybe(
					h, self.account, f'{self.name} from this account [{make_timestr()}]')
				if dest_addr_chk:
					wallet_data = h.get_wallet_data(print=False)
				dest_addr, dest_addr_idx = h.get_last_addr(self.account, wallet_data, display=not dest_addr_chk)
				if dest_addr_chk:
					h.print_acct_addrs(wallet_data, self.account)
			elif self.dest_acct is None: # sweep to wallet
				h.close_wallet('source')
				h2 = self.rpc(self, self.dest)
				h2.open_wallet('destination')
				wallet_data2 = h2.get_wallet_data()

				wf = self.get_wallet_fn(self.dest)
				if keypress_confirm(self.cfg, f'\nCreate new account for wallet {wf.name!r}?'):
					dest_acct, dest_addr = h2.create_acct(
						label = f'{self.name} from {self.source.idx}:{self.account} [{make_timestr()}]')
					dest_addr_idx = 0
					h2.get_wallet_data()
				elif keypress_confirm(self.cfg, f'Sweep to last existing account of wallet {wf.name!r}?'):
					dest_acct, dest_addr_chk = h2.get_last_acct(wallet_data2.accts_data)
					dest_addr, dest_addr_idx = h2.get_last_addr(dest_acct, wallet_data2, display=False)
				else:
					die(1, 'Exiting at user request')

				h2.close_wallet('destination')
				h.open_wallet('source', refresh=False)
			else: # sweep to specific account of wallet

				def get_dest_addr_params(h, wallet_data, dest_acct, label):
					self.check_account_exists(wallet_data.accts_data, dest_acct)
					h.print_acct_addrs(wallet_data, dest_acct)
					dest_addr_chk = create_new_addr_maybe(h, dest_acct, label)
					if dest_addr_chk:
						wallet_data = h.get_wallet_data(print=False)
					dest_addr, dest_addr_idx = h.get_last_addr(dest_acct, wallet_data, display=not dest_addr_chk)
					if dest_addr_chk:
						h.print_acct_addrs(wallet_data, dest_acct)
					return dest_addr, dest_addr_idx, dest_addr_chk

				dest_acct = self.dest_acct

				if self.dest == self.source:
					dest_addr, dest_addr_idx, dest_addr_chk = get_dest_addr_params(
						h, wallet_data, dest_acct,
						f'{self.name} from account #{self.account} [{make_timestr()}]')
				else:
					h.close_wallet('source')
					h2 = self.rpc(self, self.dest)
					h2.open_wallet('destination')
					dest_addr, dest_addr_idx, dest_addr_chk = get_dest_addr_params(
						h2, h2.get_wallet_data(), dest_acct,
						f'{self.name} from {self.source.idx}:{self.account} [{make_timestr()}]')
					h2.close_wallet('destination')
					h.open_wallet('source', refresh=False)

			assert dest_addr_chk in (None, dest_addr), (
				f'dest_addr: ({dest_addr}) != dest_addr_chk: ({dest_addr_chk})')

			msg(f'\n    Creating {self.name} transaction...')
			return (h, h.make_sweep_tx(self.account, dest_acct, dest_addr_idx, dest_addr, wallet_data.addrs_data))

		@property
		def add_desc(self):
			return (
				r' to new address' if self.dest is None else
				f' to new account in wallet {self.dest.idx}' if self.dest_acct is None else
				f' to account #{self.dest_acct} of wallet {self.dest.idx}') + f' ({self.sweep_type} sweep)'

		def check_account_exists(self, accts_data, idx):
			max_acct = len(accts_data['subaddress_accounts']) - 1
			if self.account > max_acct:
				die(2, f'{self.account}: requested account index out of bounds (>{max_acct})')

		async def main(self):

			gmsg(
				f'\n{self.stem.capitalize()}ing account #{self.account}'
				f' of wallet {self.source.idx}{self.add_desc}')

			h = self.rpc(self,self.source)

			h.open_wallet('source')

			wallet_data = h.get_wallet_data(skip_empty_ok=True)

			self.check_account_exists(wallet_data.accts_data, self.account)

			h.print_acct_addrs(wallet_data, self.account)

			h, new_tx = self.create_tx(h, wallet_data)

			msg('\n' + new_tx.get_info(indent='    '))

			if self.cfg.tx_relay_daemon:
				self.display_tx_relay_info(indent='    ')

			msg('Saving TX data to file')
			new_tx.write(delete_metadata=True)

			if self.cfg.no_relay or self.cfg.autosign:
				return True

			if keypress_confirm( self.cfg, f'Relay {self.name} transaction?' ):
				if self.cfg.tx_relay_daemon:
					await h.stop_wallet('source')
					msg('')
					self.init_tx_relay_daemon()
					h = self.rpc(self,self.source)
					h.open_wallet('TX-relay-configured source',refresh=False)
				msg_r(f'\n    Relaying {self.name} transaction...')
				h.relay_tx(new_tx.data.metadata)
				gmsg('\nAll done')
				return True
			else:
				die(1,'\nExiting at user request')

	class sweep_all(sweep):
		stem = 'sweep'
		sweep_type = 'all-address'

	class transfer(sweep):
		stem    = 'transferr'
		spec_id = 'transfer_spec'
		spec_key = ( (1,'source'), )

		@property
		def add_desc(self):
			return f': {self.amount} XMR to {self.dest_addr}'

		def create_tx(self, h, wallet_data):
			msg(f'\n    Creating {self.name} transaction...')
			return (h, h.make_transfer_tx(self.account, self.dest_addr, self.amount))

	class new(spec):
		spec_id = 'newaddr_spec'
		spec_key = ( (1,'source'), )
		wallet_offline = True

		async def main(self):
			h = self.rpc(self,self.source)
			h.open_wallet('Monero')

			desc = 'account' if self.account is None else 'address'
			label = (
				None if self.label == '' else
				'{} [{}]'.format(self.label or f'xmrwallet new {desc}', make_timestr()))

			wallet_data = h.get_wallet_data()

			if desc == 'address':
				h.print_acct_addrs(wallet_data, self.account)

			if keypress_confirm(
					self.cfg,
					'\nCreating new {a} for wallet {b}{c} with {d}\nOK?'.format(
						a = desc,
						b = red(str(self.source.idx)),
						c = '' if desc == 'account' else f', account {red("#"+str(self.account))}',
						d = 'label ' + pink('‘'+label+'’') if label else 'empty label')
					):

				if desc == 'address':
					h.create_new_addr(self.account, label=label)
				else:
					h.create_acct(label=label)

				wallet_data = h.get_wallet_data(print=desc=='account')

				if desc == 'address':
					h.print_acct_addrs(wallet_data, self.account)
			else:
				ymsg('\nOperation cancelled by user request')

			# wallet must be left open: otherwise the 'stop_wallet' RPC call used to stop the daemon will fail
			if self.cfg.no_stop_wallet_daemon:
				h.close_wallet('Monero')

			msg('')

	class label(spec):
		spec_id  = 'label_spec'
		spec_key = ( (1,'source'), )
		opts     = ()
		wallet_offline = True

		async def main(self):

			gmsg('\n{a} label for wallet {b}, account #{c}, address #{d}'.format(
				a = 'Setting' if self.label else 'Removing',
				b = self.source.idx,
				c = self.account,
				d = self.address_idx
			))
			h = self.rpc(self,self.source)

			h.open_wallet('source')
			wallet_data = h.get_wallet_data()

			max_acct = len(wallet_data.accts_data['subaddress_accounts']) - 1
			if self.account > max_acct:
				die(2, f'{self.account}: requested account index out of bounds (>{max_acct})')

			ret = h.print_acct_addrs(wallet_data, self.account)

			if self.address_idx > len(ret) - 1:
				die(2, '{}: requested address index out of bounds (>{})'.format(
					self.address_idx,
					len(ret) - 1 ))

			addr = ret[self.address_idx]
			new_label = f'{self.label} [{make_timestr()}]' if self.label else ''

			ca = CoinAddr(self.proto, addr['address'])
			msg('\n  {a} {b}\n  {c} {d}\n  {e} {f}'.format(
					a = 'Address:       ',
					b = ca.hl() if self.cfg.full_address else ca.fmt(color=True, width=addr_width),
					c = 'Existing label:',
					d = pink(addr['label']) if addr['label'] else gray('[none]'),
					e = 'New label:     ',
					f = pink(new_label) if new_label else gray('[none]') ))

			op = 'remove' if not new_label else 'update' if addr['label'] else 'set'

			if addr['label'] == new_label:
				ymsg('\nLabel is unchanged, operation cancelled')
			elif keypress_confirm(self.cfg, f'  {op.capitalize()} label?'):
				h.set_label(self.account, self.address_idx, new_label)
				ret = h.print_acct_addrs(h.get_wallet_data(print=False), self.account)
				label_chk = ret[self.address_idx]['label']
				if label_chk != new_label:
					ymsg(f'Warning: new label {label_chk!r} does not match requested value!')
					return False
				else:
					msg(cyan('\nLabel successfully {}'.format('set' if op == 'set' else op+'d')))
			else:
				ymsg('\nOperation cancelled by user request')

	class sign(wallet):
		action = 'signing transaction with'
		start_daemon = False

		async def main(self,fn,restart_daemon=True):
			if restart_daemon:
				await self.restart_wallet_daemon()
			tx = MoneroMMGenTX.Unsigned( self.cfg, fn )
			h = self.rpc(self,self.addr_data[0])
			self.head_msg(tx.src_wallet_idx,h.fn)
			if restart_daemon:
				h.open_wallet(refresh=False)
			res = self.c.call(
				'sign_transfer',
				unsigned_txset = tx.data.unsigned_txset,
				export_raw = True,
				get_tx_keys = True
			)
			new_tx = MoneroMMGenTX.NewColdSigned(
				cfg            = self.cfg,
				txid           = res['tx_hash_list'][0],
				unsigned_txset = None,
				signed_txset   = res['signed_txset'],
				_in_tx         = tx,
			)
			return new_tx

	class submit(wallet):
		action = 'submitting transaction with'
		opts = ('tx_relay_daemon',)

		def post_mount_action(self):
			self.tx # trigger an exit if no suitable transaction present

		@property
		def tx(self):
			if not hasattr(self,'_tx'):
				self._tx = self.get_tx()
			return self._tx

		def get_tx(self):
			if uarg.infile:
				fn = Path(uarg.infile)
			else:
				from .autosign import Signable
				fn = Signable.xmr_transaction(self.asi).get_unsubmitted()
			return MoneroMMGenTX.ColdSigned(cfg=self.cfg, fn=fn)

		def get_relay_rpc(self):

			relay_opt = self.parse_tx_relay_opt()

			wd = MoneroWalletDaemon(
				cfg         = self.cfg,
				proto       = self.proto,
				wallet_dir  = self.cfg.wallet_dir or '.',
				test_suite  = self.cfg.test_suite,
				monerod_addr = relay_opt[1],
			)

			u = wd.usr_daemon_args = []
			if self.cfg.test_suite:
				u.append('--daemon-ssl-allow-any-cert')
			if relay_opt[2]:
				u.append(f'--proxy={relay_opt[2]}')

			return MoneroWalletRPCClient(
				cfg             = self.cfg,
				daemon          = wd,
				test_connection = False,
			)

		async def main(self):
			tx = self.tx
			h = self.rpc( self, self.kal.entry(tx.src_wallet_idx) )
			self.head_msg(tx.src_wallet_idx,h.fn)
			h.open_wallet()

			if self.cfg.tx_relay_daemon:
				await self.c.stop_daemon()
				self.c = self.get_relay_rpc()
				self.c.start_daemon()
				h = self.rpc( self, self.kal.entry(tx.src_wallet_idx) )
				h.open_wallet( 'TX-relay-configured watch-only', refresh=False )

			msg('\n' + tx.get_info(indent='    '))

			if self.cfg.tx_relay_daemon:
				self.display_tx_relay_info(indent='    ')

			if keypress_confirm( self.cfg, f'{self.name.capitalize()} transaction?' ):
				if self.cfg.tx_relay_daemon:
					msg_r('Relaying transaction to remote daemon, please be patient...')
					t_start = time.time()
				res = self.c.call(
					'submit_transfer',
					tx_data_hex = tx.data.signed_txset )
				assert res['tx_hash_list'][0] == tx.data.txid, 'TxID mismatch in ‘submit_transfer’ result!'
				if self.cfg.tx_relay_daemon:
					from .util2 import format_elapsed_hr
					msg(f'success\nRelay time: {format_elapsed_hr(t_start, rel_now=False, show_secs=True)}')
			else:
				die(1,'Exiting at user request')

			new_tx = MoneroMMGenTX.NewSubmitted(
				cfg          = self.cfg,
				_in_tx       = tx,
			)
			gmsg('\nOK')
			new_tx.write(
				ask_write     = not self.cfg.autosign,
				ask_overwrite = not self.cfg.autosign )
			return new_tx

	class resubmit(submit):
		action = 'resubmitting transaction with'

		def check_uopts(self):
			if not self.cfg.autosign:
				die(1,'--autosign is required for this operation')

		def get_tx(self):
			from .autosign import Signable
			fns = Signable.xmr_transaction(self.asi).get_submitted()
			return sorted(
				(MoneroMMGenTX.Submitted(self.cfg, Path(fn)) for fn in fns),
					key = lambda x: getattr(x.data,'submit_time',None) or x.data.create_time
			)[-1]

	class abort(base):
		opts = ('watch_only','autosign')

		def __init__(self, cfg, uarg_tuple):
			super().__init__(cfg,uarg_tuple)
			self.mount_removable_device()
			from .autosign import Signable
			Signable.xmr_transaction(self.asi).shred_abortable() # prompts user, then raises exception or exits

	class dump(wallet):
		wallet_offline = True

		async def process_wallet(self,d,fn,last):
			h = self.rpc(self,d)
			h.open_wallet('source')
			wallet_data = h.get_wallet_data(print=False)
			msg('')
			MoneroWalletDumpFile.New(
				parent    = self,
				wallet_fn = fn,
				data      = {'wallet_metadata': wallet_data.addrs_data}
			).write()
			return True

	class export_outputs(wallet):
		action = 'exporting outputs from'
		stem = 'process'
		sign = False

		async def process_wallet(self,d,fn,last):
			h = self.rpc(self,d)
			h.open_wallet('source')

			if self.cfg.rescan_blockchain:
				gmsg_r('\n  Rescanning blockchain...')
				self.c.call('rescan_blockchain')
				gmsg('done')

			if self.cfg.rescan_spent:
				gmsg_r('\n  Rescanning spent outputs...')
				self.c.call('rescan_spent')
				gmsg('done')

			self.head_msg(d.idx,h.fn)
			for ftype in ('Unsigned','Signed'):
				old_fn = getattr(MoneroWalletOutputsFile,ftype).find_fn_from_wallet_fn(
					cfg             = self.cfg,
					wallet_fn       = fn,
					ret_on_no_match = True )
				if old_fn:
					old_fn.unlink()
			m = MoneroWalletOutputsFile.New(
				parent    = self,
				wallet_fn = fn,
				data      = self.c.call('export_outputs', all=True),
				sign      = self.sign,
			)
			m.write()
			return True

	class export_outputs_sign(export_outputs):
		opts = ('rescan_spent','rescan_blockchain')
		sign = True

	class import_outputs(wallet):
		action = 'importing wallet outputs into'
		start_daemon = False

		async def main(self,fn,wallet_idx,restart_daemon=True):
			if restart_daemon:
				await self.restart_wallet_daemon()
			h = self.rpc(self,self.addr_data[0])
			self.head_msg(wallet_idx,fn)
			if restart_daemon:
				h.open_wallet(refresh=False)
			m = MoneroWalletOutputsFile.Unsigned(
				parent = self,
				fn     = fn )
			res = self.c.call(
				'import_outputs',
				outputs_data_hex = m.data.outputs_data_hex )
			idata = res['num_imported']
			bmsg(f'\n  {idata} output{suf(idata)} imported')
			if m.data.sign:
				data = m.data._asdict()
				data.update(self.c.call('export_key_images', all=True))
				m = MoneroWalletOutputsFile.SignedNew(
					parent    = self,
					wallet_fn = m.get_wallet_fn(fn),
					data      = data)
				idata = m.data.signed_key_images or []
				bmsg(f'  {len(idata)} key image{suf(idata)} signed')
			else:
				m.data = m.data._replace(imported=True)
			return m

	class import_key_images(wallet):
		action = 'importing key images into'
		stem = 'process'
		trust_monerod = True

		def post_main_failure(self):
			rw_msg = ' for requested wallets' if uarg.wallets else ''
			die(2, f'No signed key image files found{rw_msg}!')

		async def process_wallet(self,d,fn,last):
			keyimage_fn = MoneroWalletOutputsFile.Signed.find_fn_from_wallet_fn( self.cfg, fn, ret_on_no_match=True )
			if not keyimage_fn:
				msg(f'No signed key image file found for wallet #{d.idx}')
				return False
			h = self.rpc(self,d)
			h.open_wallet()
			self.head_msg(d.idx,h.fn)
			m = MoneroWalletOutputsFile.Signed( parent=self, fn=keyimage_fn )
			data = m.data.signed_key_images or []
			bmsg(f'\n  {len(data)} signed key image{suf(data)} to import')
			if data:
				res = self.c.call( 'import_key_images', signed_key_images=data )
				bmsg(f'  Success: {res}')
			return True

	class relay(base):
		opts = ('tx_relay_daemon',)

		def __init__(self,cfg,uarg_tuple):

			super().__init__(cfg,uarg_tuple)

			self.mount_removable_device()

			self.tx = MoneroMMGenTX.Signed( self.cfg, Path(uarg.infile) )

			if self.cfg.tx_relay_daemon:
				m = self.parse_tx_relay_opt()
				host,port = m[1].split(':')
				proxy = m[2]
				md = None
			else:
				from .daemon import CoinDaemon
				md = CoinDaemon( self.cfg, 'xmr', test_suite=self.cfg.test_suite )
				host,port = ('localhost', md.rpc_port)
				proxy = None

			self.dc = MoneroRPCClient(
				cfg    = self.cfg,
				proto  = self.proto,
				daemon = md,
				host   = host,
				port   = int(port),
				user   = None,
				passwd = None,
				test_connection = host == 'localhost', # avoid extra connections if relay is a public node
				proxy  = proxy )

		async def main(self):
			msg('\n' + self.tx.get_info(indent='    '))

			if self.cfg.tx_relay_daemon:
				self.display_tx_relay_info(indent='    ')

			if keypress_confirm( self.cfg, 'Relay transaction?' ):
				if self.cfg.tx_relay_daemon:
					msg_r('Relaying transaction to remote daemon, please be patient...')
					t_start = time.time()
				res = self.dc.call_raw(
					'send_raw_transaction',
					tx_as_hex = self.tx.data.blob
				)
				if res['status'] == 'OK':
					if res['not_relayed']:
						msg('not relayed')
						ymsg('Transaction not relayed')
					else:
						msg('success')
					if self.cfg.tx_relay_daemon:
						from .util2 import format_elapsed_hr
						msg(f'Relay time: {format_elapsed_hr(t_start, rel_now=False, show_secs=True)}')
					gmsg('OK')
					return True
				else:
					die( 'RPCFailure', repr(res) )
			else:
				die(1,'Exiting at user request')

	class txview(base):
		view_method = 'get_info'
		opts = ('watch_only','autosign')
		hdr = ''
		col_hdr = ''
		footer = ''
		do_umount = False

		async def main(self, cols=None):

			self.mount_removable_device()

			if self.cfg.autosign:
				files = [f for f in self.asi.xmr_tx_dir.iterdir()
							if f.name.endswith('.'+MoneroMMGenTX.Submitted.ext)]
			else:
				files = uarg.infile

			txs = sorted(
				(MoneroMMGenTX.View( self.cfg, Path(fn) ) for fn in files),
					# old TX files have no ‘submit_time’ field:
					key = lambda x: getattr(x.data,'submit_time',None) or x.data.create_time
			)

			if self.cfg.autosign:
				self.asi.do_umount()

			addr_w = None if self.cfg.full_address or cols is None else cols - self.fixed_cols_w

			self.cfg._util.stdout_or_pager(
				(self.hdr if len(files) > 1 else '')
				+ self.col_hdr
				+ '\n'.join(getattr(tx, self.view_method)(addr_w=addr_w) for tx in txs)
				+ self.footer
			)

	class txlist(txview):
		view_method = 'get_info_oneline'
		add_nl = True
		footer = '\n'
		fixed_cols_w = MoneroMMGenTX.Base.oneline_fixed_cols_w
		min_addr_w = 10

		@property
		def hdr(self):
			return ('SUBMITTED ' if self.cfg.autosign else '') + 'MONERO TRANSACTIONS\n'

		@property
		def col_hdr(self):
			return MoneroMMGenTX.View.oneline_fs.format(
				a = 'Network',
				b = 'Seed ID',
				c = 'Submitted' if self.cfg.autosign else 'Date',
				d = 'TxID',
				e = 'Type',
				f = 'Src',
				g = 'Dest',
				h = '  Amount',
				j = 'Dest Address',
				x = '',
			) + '\n'

		async def main(self):
			if self.cfg.pager:
				cols = None
			else:
				from .term import get_terminal_size
				cols = self.cfg.columns or get_terminal_size().width
				if cols < self.fixed_cols_w + self.min_addr_w:
					die(1, f'A terminal at least {self.fixed_cols_w + self.min_addr_w} columns wide is required '
							'to display this output (or use --columns or --pager)' )
			await super().main(cols=cols)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_py_d.ct_wallet: Wallet conversion tests for the cmdtest.py test suite
"""

import os

from mmgen.util import msg,capfirst,get_extension
from mmgen.wallet import get_wallet_cls

from ..include.common import cfg,joinpath,imsg
from .common import ref_dir,ref_wallet_brainpass,ref_wallet_incog_offset,hincog_fn,hincog_bytes
from .ct_base import CmdTestBase
from .ct_shared import CmdTestShared

class CmdTestWalletConv(CmdTestBase,CmdTestShared):
	'wallet conversion to and from reference data'
	networks = ('btc',)
	tmpdir_nums = [11,12,13]
	sources = { '128': {
					'ref_wallet':      'FE3C6545-D782B529[128,1].mmdat',
					'ic_wallet':       'FE3C6545-E29303EA-5E229E30[128,1].mmincog',
					'ic_wallet_hex':   'FE3C6545-BC4BE3F2-32586837[128,1].mmincox',

					'hic_wallet':       'FE3C6545-161E495F-BEB7548E[128,1].incog-offset123',
					'hic_wallet_old':   'FE3C6545-161E495F-9860A85B[128,1].incog-old.offset123',
				},
				'192': {
					'ref_wallet':      '1378FC64-6F0F9BB4[192,1].mmdat',
					'ic_wallet':       '1378FC64-2907DE97-F980D21F[192,1].mmincog',
					'ic_wallet_hex':   '1378FC64-4DCB5174-872806A7[192,1].mmincox',

					'hic_wallet':      '1378FC64-B55E9958-77256FC1[192,1].incog.offset123',
					'hic_wallet_old':  '1378FC64-B55E9958-D85FF20C[192,1].incog-old.offset123',
				},
				'256': {
					'ref_wallet':      '98831F3A-27F2BF93[256,1].mmdat',
					'ic_wallet':       '98831F3A-5482381C-18460FB1[256,1].mmincog',
					'ic_wallet_hex':   '98831F3A-1630A9F2-870376A9[256,1].mmincox',

					'hic_wallet':       '98831F3A-F59B07A0-559CEF19[256,1].incog.offset123',
					'hic_wallet_old':   '98831F3A-F59B07A0-848535F3[256,1].incog-old.offset123',

				},
			}
	cmd_group = (
		# reading
		('ref_wallet_conv',    'conversion of saved reference wallet'),
		('ref_mn_conv',        'conversion of saved MMGen native mnemonic'),
		('ref_bip39_conv',     'conversion of saved BIP39 mnemonic'),
		('ref_seed_conv',      'conversion of saved seed file'),
		('ref_hex_conv',       'conversion of saved MMGen hexadecimal seed file'),
		('ref_plainhex_conv',  'conversion of saved plain hexadecimal seed file'),
		('ref_dieroll_conv',   'conversion of saved dieroll (b6d) seed file'),
		('ref_brain_conv',     'conversion of ref brainwallet'),
		('ref_incog_conv',     'conversion of saved incog wallet'),
		('ref_incox_conv',     'conversion of saved hex incog wallet'),
		('ref_hincog_conv',    'conversion of saved hidden incog wallet'),
		('ref_hincog_conv_old','conversion of saved hidden incog wallet (old format)'),
		# writing
		('ref_wallet_conv_out', 'ref seed conversion to wallet'),
		('ref_mn_conv_out',     'ref seed conversion to MMGen native mnemonic'),
		('ref_bip39_conv_out',  'ref seed conversion to BIP39 mnemonic'),
		('ref_hex_conv_out',    'ref seed conversion to MMGen hex seed'),
		('ref_plainhex_conv_out','ref seed conversion to plain hex seed'),
		('ref_dieroll_conv_out','ref seed conversion to dieroll (b6d) seed'),
		('ref_seed_conv_out',   'ref seed conversion to seed'),
		('ref_incog_conv_out',  'ref seed conversion to incog data'),
		('ref_incox_conv_out',  'ref seed conversion to hex incog data'),
		('ref_hincog_conv_out', 'ref seed conversion to hidden incog data'),
		('ref_hincog_blkdev_conv_out', 'ref seed conversion to hidden incog data on block device')
	)

	def __init__(self,trunner,cfgs,spawn):
		for k,_ in self.cmd_group:
			for n in (1,2,3):
				setattr(self,f'{k}_{n}',getattr(self,k))
		CmdTestBase.__init__(self,trunner,cfgs,spawn)

	def ref_wallet_conv(self):
		wf = joinpath(ref_dir,self.sources[str(self.seed_len)]['ref_wallet'])
		return self.walletconv_in(wf)

	def ref_mn_conv(self,ext='mmwords'):
		wf = joinpath(ref_dir,self.seed_id+'.'+ext)
		return self.walletconv_in(wf)

	def ref_bip39_conv(self):
		return self.ref_mn_conv(ext='bip39')
	def ref_seed_conv(self):
		return self.ref_mn_conv(ext='mmseed')
	def ref_hex_conv(self):
		return self.ref_mn_conv(ext='mmhex')
	def ref_plainhex_conv(self):
		return self.ref_mn_conv(ext='hex')
	def ref_dieroll_conv(self):
		return self.ref_mn_conv(ext='b6d')

	def ref_brain_conv(self):
		uopts = ['-i','bw','-p','1','-l',str(self.seed_len)]
		return self.walletconv_in(None,uopts,icls=get_wallet_cls('brain'))

	def ref_incog_conv(self,wfk='ic_wallet',in_fmt='i'):
		uopts = ['-i',in_fmt,'-p','1','-l',str(self.seed_len)]
		wf = joinpath(ref_dir,self.sources[str(self.seed_len)][wfk])
		return self.walletconv_in(wf,uopts)

	def ref_incox_conv(self):
		return self.ref_incog_conv(in_fmt='xi',wfk='ic_wallet_hex')

	def ref_hincog_conv(self,wfk='hic_wallet',add_uopts=[]):
		ic_f = joinpath(ref_dir,self.sources[str(self.seed_len)][wfk])
		uopts = ['-i','hi','-p','1','-l',str(self.seed_len)] + add_uopts
		hi_opt = ['-H',f'{ic_f},{ref_wallet_incog_offset}']
		return self.walletconv_in(
			None,
			uopts + hi_opt,
			icls = get_wallet_cls('incog_hidden') )

	def ref_hincog_conv_old(self):
		return self.ref_hincog_conv(wfk='hic_wallet_old',add_uopts=['-O'])

	def ref_wallet_conv_out(self):
		return self.walletconv_out('w')
	def ref_mn_conv_out(self):
		return self.walletconv_out('mn')
	def ref_bip39_conv_out(self):
		return self.walletconv_out('bip39')
	def ref_seed_conv_out(self):
		return self.walletconv_out('seed')
	def ref_hex_conv_out(self):
		return self.walletconv_out('hexseed')
	def ref_plainhex_conv_out(self):
		return self.walletconv_out('hex')
	def ref_dieroll_conv_out(self):
		return self.walletconv_out('dieroll')
	def ref_incog_conv_out(self):
		return self.walletconv_out('i')
	def ref_incox_conv_out(self):
		return self.walletconv_out('xi')

	def ref_hincog_conv_out(self,ic_f=None):
		if not ic_f:
			ic_f = joinpath(self.tmpdir,hincog_fn)
		hi_parms = f'{ic_f},{ref_wallet_incog_offset}'
		sl_parm = '-l' + str(self.seed_len)
		return self.walletconv_out('hi',
									uopts     = ['-J',hi_parms,sl_parm],
									uopts_chk = ['-H',hi_parms,sl_parm] )

	def ref_hincog_blkdev_conv_out(self):
		def do_run(cmd):
			from subprocess import run,PIPE,DEVNULL
			return run(cmd,stdout=PIPE,stderr=DEVNULL,check=True)
		if self.skip_for_win():
			return 'skip'
		imsg('Creating block device image file')
		ic_img = joinpath(self.tmpdir,'hincog_blkdev_img')
		do_run(['dd','if=/dev/zero','of='+ic_img,'bs=1K','count=1'])
		ic_dev = do_run(['sudo','/sbin/losetup','-f']).stdout.strip().decode()
		ic_dev_mode_orig = '{:o}'.format( os.stat(ic_dev).st_mode & 0xfff )
		ic_dev_mode = '0666'
		imsg(f'Changing permissions on loop device to {ic_dev_mode!r}')
		do_run(['sudo','chmod',ic_dev_mode,ic_dev])
		imsg(f'Attaching loop device {ic_dev!r}')
		do_run(['sudo','/sbin/losetup',ic_dev,ic_img])
		self.ref_hincog_conv_out(ic_f=ic_dev)
		imsg(f'Detaching loop device {ic_dev!r}')
		do_run(['sudo','/sbin/losetup','-d',ic_dev])
		imsg(f'Resetting permissions on loop device to {ic_dev_mode_orig!r}')
		do_run(['sudo','chmod',ic_dev_mode_orig,ic_dev])
		return 'ok'

	# wallet conversion tests
	def walletconv_in(self,infile,uopts=[],icls=None):
		ocls = get_wallet_cls('words')
		opts = ['-d',self.tmpdir,'-o',ocls.fmt_codes[0],self.usr_rand_arg]
		if_arg = [infile] if infile else []
		d = '(convert)'
		t = self.spawn('mmgen-walletconv',opts+uopts+if_arg,extra_desc=d)
		t.license()
		icls = icls or get_wallet_cls(ext=get_extension(infile))
		if icls.type == 'brain':
			t.expect('Enter brainwallet: ',ref_wallet_brainpass+'\n')
		if icls.enc and icls.type != 'brain':
			t.passphrase(icls.desc,self.wpasswd)
			if self.test_name[:19] == 'ref_hincog_conv_old':
				t.expect('Is the Seed ID correct? (Y/n): ','\n')
			else:
				t.expect(['Passphrase is OK',' are correct'])
		wf = t.written_to_file(capfirst(ocls.desc))
		t.p.wait()
		# back check of result
		msg('' if cfg.profile else ' OK')
		return self.walletchk(
				wf,
				extra_desc = '(check)',
				sid        = self.seed_id)

	def walletconv_out(self,out_fmt='w',uopts=[],uopts_chk=[]):
		wcls = get_wallet_cls(fmt_code=out_fmt)
		opts = ['-d',self.tmpdir,'-p1','-o',out_fmt] + uopts
		infile = joinpath(ref_dir,self.seed_id+'.mmwords')
		t = self.spawn('mmgen-walletconv',[self.usr_rand_arg]+opts+[infile],extra_desc='(convert)')

		add_args = [f'-l{self.seed_len}']
		t.license()
		if wcls.enc and wcls.type != 'brain':
			t.passphrase_new('new '+wcls.desc,self.wpasswd)
			t.usr_rand(self.usr_rand_chars)
		if wcls.type.startswith('incog'):
			for _ in range(3):
				t.expect('Encrypting random data from your operating system with ephemeral key')
		if wcls.type == 'incog_hidden':
			t.hincog_create(hincog_bytes)
		if out_fmt == 'w':
			t.label()
		wf = t.written_to_file(capfirst(wcls.desc))

		if wcls.type == 'incog_hidden':
			add_args += uopts_chk
			wf = None
		msg('' if cfg.profile else ' OK')
		return self.walletchk(
				wf,
				wcls       = wcls,
				extra_desc = '(check)',
				sid        = self.seed_id,
				add_args   = add_args)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
help: help notes for MMGen suite commands
"""

from ..cfg import gc

def help_notes_func(proto,cfg,k):

	def fee_spec_letters(use_quotes=False):
		cu = proto.coin_amt.units
		sep,conj = ((',',' or '),("','","' or '"))[use_quotes]
		return sep.join(u[0] for u in cu[:-1]) + ('',conj)[len(cu)>1] + cu[-1][0]

	def fee_spec_names():
		cu = proto.coin_amt.units
		return ', '.join(cu[:-1]) + ('',' and ')[len(cu)>1] + cu[-1] + ('',',\nrespectively')[len(cu)>1]

	def coind_exec():
		from ..daemon import CoinDaemon
		return (
			CoinDaemon(cfg,proto.coin).exec_fn if proto.coin in CoinDaemon.coins else 'bitcoind' )

	class help_notes:

		def dfl_twname():
			from ..proto.btc.rpc import BitcoinRPCClient
			return BitcoinRPCClient.dfl_twname

		def MasterShareIdx():
			from ..seedsplit import MasterShareIdx
			return MasterShareIdx

		def tool_help():
			from ..tool.help import main_help
			return main_help()

		def dfl_subseeds():
			from ..subseed import SubSeedList
			return str(SubSeedList.dfl_len)

		def dfl_seed_len():
			from ..seed import Seed
			return str(Seed.dfl_len)

		def password_formats():
			from ..passwdlist import PasswordList
			pwi_fs = '{:8} {:1} {:26} {:<7}  {:<7}  {}'
			return '\n  '.join(
				[pwi_fs.format('Code','','Description','Min Len','Max Len','Default Len')] +
				[pwi_fs.format(k,'-',v.desc,v.min_len,v.max_len,v.dfl_len) for k,v in PasswordList.pw_info.items()]
			)

		def dfl_mmtype():
			from ..addr import MMGenAddrType
			return "'{}' or '{}'".format(
				proto.dfl_mmtype,
				MMGenAddrType.mmtypes[proto.dfl_mmtype].name )

		def address_types():
			from ..addr import MMGenAddrType
			return '\n  '.join([
				"'{}','{:<12} - {}".format( k, v.name+"'", v.desc )
					for k,v in MMGenAddrType.mmtypes.items()
			])

		def fmt_codes():
			from ..wallet import format_fmt_codes
			return '\n  '.join( format_fmt_codes().splitlines() )

		def coin_id():
			return proto.coin_id

		def keygen_backends():
			from ..keygen import get_backends
			from ..addr import MMGenAddrType
			backends = get_backends(
				MMGenAddrType(proto,cfg.type or proto.dfl_mmtype).pubkey_type
			)
			return ' '.join( f'{n}:{k}{" [default]" if n==1 else ""}' for n,k in enumerate(backends,1) )

		def coind_exec():
			return coind_exec()

		def coin_daemon_network_ids():
			from ..daemon import CoinDaemon
			from ..util import fmt_list
			return fmt_list(CoinDaemon.get_network_ids(cfg),fmt='bare')

		def rel_fee_desc():
			from ..tx import BaseTX
			return BaseTX(cfg=cfg,proto=proto).rel_fee_desc

		def fee_spec_letters():
			return fee_spec_letters()

		def fee():
			from ..tx import BaseTX
			return """
                               FEE SPECIFICATION

Transaction fees, both on the command line and at the interactive prompt, may
be specified as either absolute {c} amounts, using a plain decimal number, or
as {r}, using an integer followed by '{l}', for {u}.
""".format(
	c = proto.coin,
	r = BaseTX(cfg=cfg,proto=proto).rel_fee_desc,
	l = fee_spec_letters(use_quotes=True),
	u = fee_spec_names() )

		def passwd():
			return """
PASSPHRASE NOTE:

For passphrases all combinations of whitespace are equal, and leading and
trailing space are ignored.  This permits reading passphrase or brainwallet
data from a multi-line file with free spacing and indentation.
""".strip()

		def brainwallet():
			return """
BRAINWALLET NOTE:

To thwart dictionary attacks, it’s recommended to use a strong hash preset
with brainwallets.  For a brainwallet passphrase to generate the correct
seed, the same seed length and hash preset parameters must always be used.
""".strip()

		def txcreate_examples():

			mmtype = 'S' if 'segwit' in proto.caps else 'C'
			from ..tool.coin import tool_cmd
			t = tool_cmd(cfg,mmtype=mmtype)
			sample_addr = t.privhex2addr('bead'*16)

			return f"""
EXAMPLES:

  Send 0.123 {proto.coin} to an external {proto.name} address, returning the change to a
  specific MMGen address in the tracking wallet:

    $ {gc.prog_name} {sample_addr},0.123 01ABCDEF:{mmtype}:7

  Same as above, but select the change address automatically:

    $ {gc.prog_name} {sample_addr},0.123 01ABCDEF:{mmtype}

  Same as above, but select the change address automatically by address type:

    $ {gc.prog_name} {sample_addr},0.123 {mmtype}

  Same as above, but reduce verbosity and specify fee of 20 satoshis
  per byte:

    $ {gc.prog_name} -q -f 20s {sample_addr},0.123 {mmtype}

  Send entire balance of selected inputs minus fee to an external {proto.name}
  address:

    $ {gc.prog_name} {sample_addr}

  Send entire balance of selected inputs minus fee to first unused wallet
  address of specified type:

    $ {gc.prog_name} {mmtype}
"""

		def txcreate():
			return f"""
The transaction’s outputs are listed on the command line, while its inputs
are chosen from a list of the wallet’s unspent outputs via an interactive
menu.  Alternatively, inputs may be specified using the --inputs option.

All addresses on the command line can be either {proto.name} addresses or MMGen
IDs in the form <seed ID>:<address type letter>:<index>.

Outputs are specified in the form <address>,<amount>, with the change output
specified by address only.  Alternatively, the change output may be an
addrlist ID in the form <seed ID>:<address type letter>, in which case the
first unused address in the tracking wallet matching the requested ID will
be automatically selected as the change output.

If the transaction fee is not specified on the command line (see FEE
SPECIFICATION below), it will be calculated dynamically using network fee
estimation for the default (or user-specified) number of confirmations.
If network fee estimation fails, the user will be prompted for a fee.

Network-estimated fees will be multiplied by the value of --fee-adjust, if
specified.

To send the value of all inputs (minus TX fee) to a single output, specify
a single address with no amount on the command line.  Alternatively, an
addrlist ID may be specified, and the address will be chosen automatically
as described above for the change output.
"""

		def txsign():
			from ..proto.btc.params import mainnet
			return """
Transactions may contain both {pnm} or non-{pnm} input addresses.

To sign non-{pnm} inputs, a {wd}flat key list is used
as the key source (--keys-from-file option).

To sign {pnm} inputs, key data is generated from a seed as with the
{pnl}-addrgen and {pnl}-keygen commands.  Alternatively, a key-address file
may be used (--mmgen-keys-from-file option).

Multiple wallets or other seed files can be listed on the command line in
any order.  If the seeds required to sign the transaction’s inputs are not
found in these files (or in the default wallet), the user will be prompted
for seed data interactively.

To prevent an attacker from crafting transactions with bogus {pnm}-to-{pnu}
address mappings, all outputs to {pnm} addresses are verified with a seed
source.  Therefore, seed files or a key-address file for all {pnm} outputs
must also be supplied on the command line if the data can’t be found in the
default wallet.
""".format(
	wd  = (f'{coind_exec()} wallet dump or ' if isinstance(proto,mainnet) else ''),
	pnm = gc.proj_name,
	pnu = proto.name,
	pnl = gc.proj_name.lower() )

		def subwallet():
			from ..subseed import SubSeedIdxRange
			return f"""
SUBWALLETS:

Subwallets (subseeds) are specified by a ‘Subseed Index’ consisting of:

  a) an integer in the range 1-{SubSeedIdxRange.max_idx}, plus
  b) an optional single letter, ‘L’ or ‘S’

The letter designates the length of the subseed.  If omitted, ‘L’ is assumed.

Long (‘L’) subseeds are the same length as their parent wallet’s seed
(typically 256 bits), while short (‘S’) subseeds are always 128-bit.
The long and short subseeds for a given index are derived independently,
so both may be used.

MMGen Wallet has no notion of ‘depth’, and to an outside observer subwallets
are identical to ordinary wallets.  This is a feature rather than a bug, as
it denies an attacker any way of knowing whether a given wallet has a parent.

Since subwallets are just wallets, they may be used to generate other
subwallets, leading to hierarchies of arbitrary depth.  However, this is
inadvisable in practice for two reasons:  Firstly, it creates accounting
complexity, requiring the user to independently keep track of a derivation
tree.  More importantly, however, it leads to the danger of Seed ID
collisions between subseeds at different levels of the hierarchy, as
MMGen checks and avoids ID collisions only among sibling subseeds.

An exception to this caveat would be a multi-user setup where sibling
subwallets are distributed to different users as their default wallets.
Since the subseeds derived from these subwallets are private to each user,
Seed ID collisions among them doesn’t present a problem.

A safe rule of thumb, therefore, is for *each user* to derive all of his/her
subwallets from a single parent.  This leaves each user with a total of two
million subwallets, which should be enough for most practical purposes.
""".strip()

	return getattr(help_notes,k)()



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
addrlist: Address list classes for the MMGen suite
"""

from .util import suf,make_chksum_N,Msg,die
from .objmethods import MMGenObject,HiliteStr,InitErrors
from .obj import MMGenListItem,ListItemAttr,MMGenDict,TwComment,WalletPassword
from .key import PrivKey
from .addr import MMGenID,MMGenAddrType,CoinAddr,AddrIdx,AddrListID,ViewKey

class AddrIdxList(tuple,InitErrors,MMGenObject):

	max_len = 1000000

	def __new__(cls,fmt_str=None,idx_list=None,sep=','):
		try:
			if fmt_str:
				def gen():
					for i in (fmt_str.split(sep)):
						j = [int(x) for x in i.split('-')]
						if len(j) == 1:
							yield j[0]
						elif len(j) == 2:
							if j[0] > j[1]:
								raise ValueError(f'{i}: invalid range')
							for k in range(j[0], j[1] + 1):
								yield k
						else:
							raise ValueError(f'{i}: invalid range')
				idx_list = tuple(gen())
			return tuple.__new__(cls,sorted({AddrIdx(i) for i in (idx_list or [])}))
		except Exception as e:
			return cls.init_fail(e,idx_list or fmt_str)

	@property
	def id_str(self):

		def gen():
			i_save = self[0]
			yield f'{i_save}'
			in_range = False

			for i in self[1:]:
				if i - i_save == 1:
					in_range = True
				else:
					if in_range:
						in_range = False
						yield f'-{i_save}'
					yield f',{i}'
				i_save = i

			if in_range:
				yield f'-{i_save}'

		return ''.join(gen()) if self else ''

class AddrListEntryBase(MMGenListItem):
	invalid_attrs = {'proto'}
	def __init__(self,proto,**kwargs):
		self.__dict__['proto'] = proto
		MMGenListItem.__init__(self,**kwargs)

class AddrListEntry(AddrListEntryBase):
	addr          = ListItemAttr(CoinAddr,include_proto=True)
	addr_p2pkh    = ListItemAttr(CoinAddr,include_proto=True)
	idx           = ListItemAttr(AddrIdx) # not present in flat addrlists
	comment       = ListItemAttr(TwComment,reassign_ok=True)
	sec           = ListItemAttr(PrivKey,include_proto=True)
	viewkey       = ListItemAttr(ViewKey,include_proto=True)
	wallet_passwd = ListItemAttr(WalletPassword)

class AddrListChksum(HiliteStr):
	color = 'pink'
	trunc_ok = False

	def __new__(cls,addrlist):
		ea = addrlist.al_id.mmtype.extra_attrs or () # add viewkey and passwd to the mix, if present
		lines = [' '.join(
					addrlist.chksum_rec_f(e) +
					tuple(getattr(e,a) for a in ea if getattr(e,a))
				) for e in addrlist.data]
		return str.__new__(cls,make_chksum_N(' '.join(lines), nchars=16, sep=True))

class AddrListIDStr(HiliteStr):
	color = 'green'
	trunc_ok = False

	def __new__(cls,addrlist,fmt_str=None):
		idxs = [e.idx for e in addrlist.data]
		prev = idxs[0]
		ret = [prev]
		for i in idxs[1:]:
			if i == prev + 1:
				if i == idxs[-1]:
					ret.extend(['-', i])
			else:
				if prev != ret[-1]:
					ret.extend(['-', prev])
				ret.extend([',', i])
			prev = i
		s = ''.join(map(str,ret))

		if fmt_str:
			ret = fmt_str.format(s)
		else:
			bc = (addrlist.proto.base_coin,addrlist.proto.coin)[addrlist.proto.base_coin=='ETH']
			mt = addrlist.al_id.mmtype
			ret = '{}{}{}[{}]'.format(
				addrlist.al_id.sid,
				('-'+bc,'')[bc == 'BTC'],
				('-'+mt,'')[mt in ('L','E')],
				s )

		addrlist.dmsg_sc('id_str',ret[8:].split('[')[0])

		return str.__new__(cls,ret)

class AddrListData(list,MMGenObject):
	pass

class AddrList(MMGenObject): # Address info for a single seed ID
	entry_type   = AddrListEntry
	main_attr    = 'addr'
	desc         = 'address'
	gen_desc     = 'address'
	gen_desc_pl  = 'es'
	gen_addrs    = True
	gen_passwds  = False
	gen_keys     = False
	has_keys     = False
	chksum_rec_f = lambda foo,e: ( str(e.idx), e.addr )

	def dmsg_sc(self,desc,data):
		Msg(f'sc_debug_{desc}: {data}')

	def noop(self,desc,data):
		pass

	def __init__(
			self,
			cfg,
			proto,
			addrfile  = '',
			al_id     = '',
			adata     = [],
			seed      = '',
			addr_idxs = '',
			src       = '',
			addrlist  = '',
			keylist   = '',
			mmtype    = None,
			key_address_validity_check = None, # None=prompt user, True=check without prompt, False=skip check
			skip_chksum = False,
			skip_chksum_msg = False,
			add_p2pkh = False ):

		self.cfg = cfg
		self.ka_validity_chk = key_address_validity_check
		self.add_p2pkh = add_p2pkh
		self.proto = proto
		do_chksum = False

		if not cfg.debug_addrlist:
			self.dmsg_sc = self.noop

		if seed and addr_idxs:   # data from seed + idxs
			self.al_id = AddrListID( sid=seed.sid, mmtype=MMGenAddrType(proto, mmtype or proto.dfl_mmtype) )
			src = 'gen'
			adata = self.generate(seed, addr_idxs if isinstance(addr_idxs,AddrIdxList) else AddrIdxList(addr_idxs))
			do_chksum = True
		elif addrfile:           # data from MMGen address file
			self.infile = addrfile
			adata = self.file.parse_file(addrfile) # sets self.al_id
			do_chksum = True
		elif al_id and adata:    # data from tracking wallet
			self.al_id = al_id
		elif addrlist:           # data from flat address list
			self.al_id = None
			from .util import remove_dups
			addrlist = remove_dups(addrlist,edesc='address',desc='address list')
			adata = AddrListData([AddrListEntry(proto=proto,addr=a) for a in addrlist])
		elif keylist:            # data from flat key list
			self.al_id = None
			from .util import remove_dups
			keylist = remove_dups(keylist,edesc='key',desc='key list',hide=True)
			adata = AddrListData([AddrListEntry(proto=proto,sec=PrivKey(proto=proto,wif=k)) for k in keylist])
		elif seed or addr_idxs:
			die(3,'Must specify both seed and addr indexes')
		elif al_id or adata:
			die(3,'Must specify both al_id and adata')
		else:
			die(3,f'Incorrect arguments for {type(self).__name__}')

		# al_id,adata now set
		self.data = adata
		self.num_addrs = len(adata)
		self.fmt_data = ''
		self.chksum = None

		if self.al_id is None:
			return

		if type(self) is ViewKeyAddrList:
			if not 'viewkey' in self.al_id.mmtype.extra_attrs:
				die(1,f'viewkeys not supported for address type {self.al_id.mmtype.desc!r}')

		self.id_str = AddrListIDStr(self)

		if type(self) is KeyList:
			return

		if do_chksum and not skip_chksum:
			self.chksum = AddrListChksum(self)
			if not skip_chksum_msg:
				self.do_chksum_msg(record=src=='gen')

	def do_chksum_msg(self,record):
		chk = 'Check this value against your records'
		rec = f'Record this checksum: it will be used to verify the {self.desc} file in the future'
		self.cfg._util.qmsg(
			f'Checksum for {self.desc} data {self.id_str.hl()}: {self.chksum.hl()}\n' +
			(chk,rec)[record] )

	def generate(self,seed,addr_idxs):

		seed = self.scramble_seed(seed.data)
		self.dmsg_sc('seed',seed[:8].hex())

		mmtype = self.al_id.mmtype

		gen_wallet_passwd = type(self) in (KeyAddrList,ViewKeyAddrList) and 'wallet_passwd' in mmtype.extra_attrs
		gen_viewkey       = type(self) in (KeyAddrList,ViewKeyAddrList) and 'viewkey' in mmtype.extra_attrs

		if self.gen_addrs:
			from .keygen import KeyGenerator
			from .addrgen import AddrGenerator
			kg = KeyGenerator( self.cfg, self.proto, mmtype.pubkey_type )
			ag = AddrGenerator( self.cfg, self.proto, mmtype )
			if self.add_p2pkh:
				ag2 = AddrGenerator( self.cfg, self.proto, 'compressed' )

		from .derive import derive_coin_privkey_bytes

		t_addrs = len(addr_idxs)
		le = self.entry_type
		out = AddrListData()
		CR = '\n' if self.cfg.debug_addrlist else '\r'

		for pk_bytes in derive_coin_privkey_bytes(seed,addr_idxs):

			if not self.cfg.debug:
				self.cfg._util.qmsg_r(
					f'{CR}Generating {self.gen_desc} #{pk_bytes.idx} ({pk_bytes.pos} of {t_addrs})' )

			e = le( proto=self.proto, idx=pk_bytes.idx )

			e.sec = PrivKey(
				self.proto,
				pk_bytes.data,
				compressed  = mmtype.compressed,
				pubkey_type = mmtype.pubkey_type )

			if self.gen_addrs:
				data = kg.gen_data(e.sec)
				e.addr = ag.to_addr(data)
				if self.add_p2pkh:
					e.addr_p2pkh = ag2.to_addr(data)
				if gen_viewkey:
					e.viewkey = ag.to_viewkey(data)
				if gen_wallet_passwd:
					e.wallet_passwd = self.gen_wallet_passwd(
						e.viewkey.encode() if type(self) is ViewKeyAddrList else e.sec )
			elif self.gen_passwds:
				e.passwd = self.gen_passwd(e.sec) # TODO - own type

			out.append(e)

		self.cfg._util.qmsg('{}{}: {} {}{} generated{}'.format(
			CR,
			self.al_id.hl(),
			t_addrs,
			self.gen_desc,
			suf(t_addrs,self.gen_desc_pl),
			' ' * 15 ))

		return out

	def gen_wallet_passwd(self,privbytes):
		from .proto.btc.common import hash256
		return WalletPassword( hash256(privbytes)[:16].hex() )

	def check_format(self,addr):
		return True # format is checked when added to list entry object

	def scramble_seed(self,seed):
		is_btcfork = self.proto.base_coin == 'BTC'
		if is_btcfork and self.al_id.mmtype == 'L' and not self.proto.testnet:
			self.dmsg_sc('str','(none)')
			return seed
		if self.proto.base_coin == 'ETH':
			scramble_key = self.proto.coin.lower()
		else:
			scramble_key = (self.proto.coin.lower()+':','')[is_btcfork] + self.al_id.mmtype.name
		from .crypto import Crypto
		if self.proto.testnet:
			scramble_key += ':' + self.proto.network
		self.dmsg_sc('str',scramble_key)
		return Crypto(self.cfg).scramble_seed(seed,scramble_key.encode())

	def idxs(self):
		return [e.idx for e in self.data]

	def addrs(self):
		return [f'{self.al_id.sid}:{e.idx}' for e in self.data]

	def addrpairs(self):
		return [(e.idx,e.addr) for e in self.data]

	def coinaddrs(self):
		return [e.addr for e in self.data]

	def comments(self):
		return [e.comment for e in self.data]

	def entry(self,idx):
		for e in self.data:
			if idx == e.idx:
				return e

	def coinaddr(self,idx):
		for e in self.data:
			if idx == e.idx:
				return e.addr

	def comment(self,idx):
		for e in self.data:
			if idx == e.idx:
				return e.comment

	def set_comment(self,idx,comment):
		for e in self.data:
			if idx == e.idx:
				e.comment = comment

	def make_reverse_dict_addrlist(self,coinaddrs):
		d = MMGenDict()
		b = coinaddrs
		for e in self.data:
			try:
				d[b[b.index(e.addr)]] = ( MMGenID(self.proto, f'{self.al_id}:{e.idx}'), e.comment )
			except ValueError:
				pass
		return d

	def add_wifs(self,key_list):
		"""
		Match WIF keys in a flat list to addresses in self by generating all
		possible addresses for each key.
		"""
		def gen_addr(pk,t):
			at = self.proto.addr_type(t)
			from .keygen import KeyGenerator
			from .addrgen import AddrGenerator
			kg = KeyGenerator( self.cfg, self.proto, at.pubkey_type )
			ag = AddrGenerator( self.cfg, self.proto, at )
			return ag.to_addr(kg.gen_data(pk))

		compressed_types = set(self.proto.mmtypes) - {'L','E'}
		uncompressed_types = set(self.proto.mmtypes) & {'L','E'}

		def gen():
			for wif in key_list:
				pk = PrivKey(proto=self.proto,wif=wif)
				for t in (compressed_types if pk.compressed else uncompressed_types):
					yield ( gen_addr(pk,t), pk )

		addrs4keys = dict(gen())

		for d in self.data:
			if d.addr in addrs4keys:
				d.sec = addrs4keys[d.addr]

	def list_missing(self,attr):
		return [d.addr for d in self.data if not getattr(d,attr)]

	@property
	def file(self):
		if not hasattr(self,'_file'):
			from . import addrfile
			self._file = getattr( addrfile, type(self).__name__.replace('List','File') )(self)
		return self._file

class KeyAddrList(AddrList):
	desc         = 'key-address'
	gen_desc     = 'key/address pair'
	gen_desc_pl  = 's'
	gen_keys     = True
	has_keys     = True
	chksum_rec_f = lambda foo,e: (str(e.idx), e.addr, e.sec.wif)

class ViewKeyAddrList(KeyAddrList):
	desc         = 'viewkey-address'
	gen_desc     = 'viewkey/address pair'
	chksum_rec_f = lambda foo,e: ( str(e.idx), e.addr )

class KeyList(KeyAddrList):
	desc         = 'key'
	gen_desc     = 'key'
	gen_addrs    = False



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.signed: Ethereum signed transaction class
"""

from ....tx import signed as TxBase
from ....obj import Str,CoinTxID,ETHNonce,HexStr
from ....addr import CoinAddr,TokenAddr
from .completed import Completed,TokenCompleted

class Signed(Completed,TxBase.Signed):

	desc = 'signed transaction'

	def parse_txfile_serialized_data(self):
		from ..pyethereum.transactions import Transaction
		from .. import rlp
		etx = rlp.decode(bytes.fromhex(self.serialized),Transaction)
		d = etx.to_dict() # ==> hex values have '0x' prefix, 0 is '0x'
		for k in ('sender','to','data'):
			if k in d:
				d[k] = d[k].replace('0x','',1)
		o = {
			'from':     CoinAddr(self.proto,d['sender']),
			# NB: for token, 'to' is token address
			'to':       CoinAddr(self.proto,d['to']) if d['to'] else Str(''),
			'amt':      self.proto.coin_amt(d['value'],'wei'),
			'gasPrice': self.proto.coin_amt(d['gasprice'],'wei'),
			'startGas': self.proto.coin_amt(d['startgas'],'wei'),
			'nonce':    ETHNonce(d['nonce']),
			'data':     HexStr(d['data']) }
		if o['data'] and not o['to']: # token- or contract-creating transaction
			# NB: could be a non-token contract address:
			o['token_addr'] = TokenAddr(self.proto,etx.creates.hex())
			self.disable_fee_check = True
		txid = CoinTxID(etx.hash.hex())
		assert txid == self.coin_txid,"txid in tx.serialized doesn't match value in MMGen transaction file"
		self.gas = o['startGas'] # approximate, but better than nothing
		self.txobj = o
		return d # 'token_addr','decimals' required by Token subclass

class TokenSigned(TokenCompleted,Signed):
	desc = 'signed transaction'

	def parse_txfile_serialized_data(self):
		raise NotImplementedError(
			'Signed transaction files cannot be parsed offline, because tracking wallet is required!')

class AutomountSigned(TxBase.AutomountSigned, Signed):
	pass

class TokenAutomountSigned(TxBase.AutomountSigned, TokenSigned):
	pass



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.new: new transaction class
"""

from collections import namedtuple

from .base import Base
from ..cfg import gc
from ..color import pink,yellow
from ..obj import get_obj,MMGenList
from ..util import msg,fmt,die,suf,remove_dups,get_extension
from ..addr import (
	is_mmgen_id,
	MMGenAddrType,
	MMGenID,
	CoinAddr,
	is_mmgen_addrtype,
	is_coin_addr,
	is_addrlist_id
)

def mmaddr2coinaddr(cfg,mmaddr,ad_w,ad_f,proto):

	def wmsg(k):
		messages = {
			'addr_in_addrfile_only': f"""
				Warning: output address {mmaddr} is not in the tracking wallet, which
				means its balance will not be tracked.  You're strongly advised to import
				the address into your tracking wallet before broadcasting this transaction.
			""",
			'addr_not_found': f"""
				No data for {gc.proj_name} address {mmaddr} could be found in either the
				tracking wallet or the supplied address file.  Please import this address
				into your tracking wallet, or supply an address file on the command line.
			""",
			'addr_not_found_no_addrfile': f"""
				No data for {gc.proj_name} address {mmaddr} could be found in the tracking
				wallet.  Please import this address into your tracking wallet or supply an
				address file for it on the command line.
			"""
		}
		return '\n' + fmt(messages[k],indent='  ')

	# assume mmaddr has already been checked
	coin_addr = ad_w.mmaddr2coinaddr(mmaddr)

	if not coin_addr:
		if ad_f:
			coin_addr = ad_f.mmaddr2coinaddr(mmaddr)
			if coin_addr:
				msg(wmsg('addr_in_addrfile_only'))
				from ..ui import keypress_confirm
				if not (cfg.yes or keypress_confirm( cfg, 'Continue anyway?' )):
					import sys
					sys.exit(1)
			else:
				die(2,wmsg('addr_not_found'))
		else:
			die(2,wmsg('addr_not_found_no_addrfile'))

	return CoinAddr(proto,coin_addr)

class New(Base):

	fee_is_approximate = False
	msg_low_coin = 'Selected outputs insufficient to fund this transaction ({} {} needed)'
	msg_wallet_low_coin = 'Wallet has insufficient funds for this transaction ({} {} needed)'
	msg_no_change_output = """
		ERROR: No change address specified.  If you wish to create a transaction with
		only one output, specify a single output address with no {} amount
	"""
	chg_autoselected = False

	def update_output_amt(self,idx,amt):
		o = self.outputs[idx]._asdict()
		o['amt'] = amt
		self.outputs[idx] = self.Output(self.proto,**o)

	def add_mmaddrs_to_outputs(self,ad_w,ad_f):
		a = [e.addr for e in self.outputs]
		d = ad_w.make_reverse_dict(a)
		if ad_f:
			d.update(ad_f.make_reverse_dict(a))
		for e in self.outputs:
			if e.addr and e.addr in d:
				e.mmid,f = d[e.addr]
				if f:
					e.comment = f

	def check_dup_addrs(self,io_str):
		assert io_str in ('inputs','outputs')
		addrs = [e.addr for e in getattr(self,io_str)]
		if len(addrs) != len(set(addrs)):
			die(2,f'{addrs}: duplicate address in transaction {io_str}')

	# given tx size and absolute fee or fee spec, return absolute fee
	# relative fee is N+<first letter of unit name>
	def feespec2abs(self,fee_arg,tx_size):
		fee = get_obj(self.proto.coin_amt,num=fee_arg,silent=True)
		if fee:
			return fee
		else:
			import re
			units = {u[0]:u for u in self.proto.coin_amt.units}
			pat = re.compile(r'([1-9][0-9]*)({})'.format('|'.join(units)))
			if pat.match(fee_arg):
				amt,unit = pat.match(fee_arg).groups()
				return self.fee_rel2abs(tx_size,units,int(amt),unit)
		return False

	def get_usr_fee_interactive(self,fee=None,desc='Starting'):
		abs_fee = None
		from ..ui import line_input
		while True:
			if fee:
				abs_fee = self.convert_and_check_fee(fee,desc)
			if abs_fee:
				prompt = '{a} TX fee{b}: {c}{d} {e} ({f} {g})\n'.format(
						a = desc,
						b = (f' (after {self.cfg.fee_adjust:.2f}X adjustment)'
							if self.cfg.fee_adjust != 1 and desc.startswith('Network-estimated')
								else ''),
						c = ('','≈')[self.fee_is_approximate],
						d = abs_fee.hl(),
						e = self.coin,
						f = pink(str(self.fee_abs2rel(abs_fee))),
						g = self.rel_fee_disp)
				from ..ui import keypress_confirm
				if self.cfg.yes or keypress_confirm( self.cfg, prompt+'OK?', default_yes=True ):
					if self.cfg.yes:
						msg(prompt)
					return abs_fee
			fee = line_input( self.cfg, self.usr_fee_prompt )
			desc = 'User-selected'

	# we don't know fee yet, so perform preliminary check with fee == 0
	async def precheck_sufficient_funds(self,inputs_sum,sel_unspent,outputs_sum):
		if self.twuo.total < outputs_sum:
			msg(self.msg_wallet_low_coin.format(outputs_sum-inputs_sum,self.dcoin))
			return False
		if inputs_sum < outputs_sum:
			msg(self.msg_low_coin.format(outputs_sum-inputs_sum,self.dcoin))
			return False
		return True

	async def get_fee_from_user(self,have_estimate_fail=[]):

		if self.cfg.fee:
			desc = 'User-selected'
			start_fee = self.cfg.fee
		else:
			desc = self.network_estimated_fee_label
			fee_per_kb,fe_type = await self.get_rel_fee_from_network()

			if fee_per_kb < 0:
				if not have_estimate_fail:
					msg(self.fee_fail_fs.format(c=self.cfg.fee_estimate_confs,t=fe_type))
					have_estimate_fail.append(True)
				start_fee = None
			else:
				start_fee = self.fee_est2abs(fee_per_kb,fe_type)

		return self.get_usr_fee_interactive(start_fee,desc=desc)

	def add_output(self,coinaddr,amt,is_chg=None):
		self.outputs.append(self.Output(self.proto,addr=coinaddr,amt=amt,is_chg=is_chg))

	def parse_cmd_arg(self, arg_in, ad_f, ad_w):

		_pa = namedtuple('parsed_txcreate_cmdline_arg', ['arg', 'mmid', 'coin_addr', 'amt'])

		arg, amt = arg_in.split(',', 1) if ',' in arg_in else (arg_in, None)

		if mmid := get_obj(MMGenID, proto=self.proto, id_str=arg, silent=True):
			coin_addr = mmaddr2coinaddr(self.cfg, arg, ad_w, ad_f, self.proto)
		elif is_coin_addr(self.proto, arg):
			coin_addr = CoinAddr(self.proto, arg)
		elif is_mmgen_addrtype(self.proto, arg) or is_addrlist_id(self.proto, arg):
			if self.proto.base_proto_coin != 'BTC':
				die(2, f'Change addresses not supported for {self.proto.name} protocol')
			self.chg_autoselected = True
			coin_addr = None
		else:
			die(2, f'{arg_in}: invalid command-line argument')

		return _pa(arg, mmid, coin_addr, amt)

	async def process_cmd_args(self,cmd_args,ad_f,ad_w):

		async def get_autochg_addr(arg, parsed_args):
			from ..tw.addresses import TwAddresses
			al = await TwAddresses(self.cfg, self.proto, get_data=True)
			exclude = [a.mmid for a in parsed_args if a.mmid]

			if is_mmgen_addrtype(self.proto, arg):
				res = al.get_change_address_by_addrtype(MMGenAddrType(self.proto, arg), exclude=exclude)
				desc = 'of address type'
			else:
				res = al.get_change_address(arg, exclude=exclude)
				desc = 'from address list'

			if res:
				return res

			die(2,'Tracking wallet contains no {t}addresses {d} {a!r}'.format(
				t = '' if res is None else 'unused ',
				d = desc,
				a = arg))

		parsed_args = [self.parse_cmd_arg(a, ad_f, ad_w) for a in cmd_args]

		chg_args = [a for a in parsed_args if not (a.amt and a.coin_addr)]

		if len(chg_args) > 1:
			desc = 'requested' if self.chg_autoselected else 'listed'
			die(2, f'ERROR: More than one change address {desc} on command line')

		for a in parsed_args:
			self.add_output(
				coinaddr = a.coin_addr or (await get_autochg_addr(a.arg, parsed_args)).addr,
				amt      = self.proto.coin_amt(a.amt or '0'),
				is_chg   = not a.amt)

		if self.chg_idx is None:
			die(2,
				fmt(self.msg_no_change_output.format(self.dcoin)).strip()
					if len(self.outputs) == 1 else
				'ERROR: No change output specified')

		if self.has_segwit_outputs() and not self.rpc.info('segwit_is_active'):
			die(2,
				f'{gc.proj_name} Segwit address requested on the command line, '
				'but Segwit is not active on this chain')

		if not self.outputs:
			die(2, 'At least one output must be specified on the command line')

	async def get_outputs_from_cmdline(self,cmd_args):
		from ..addrdata import AddrData,TwAddrData
		from ..addrlist import AddrList
		from ..addrfile import AddrFile
		addrfiles = remove_dups(
			tuple(a for a in cmd_args if get_extension(a) == AddrFile.ext),
			desc = 'command line',
			edesc = 'argument',
		)
		cmd_args  = remove_dups(
			tuple(a for a in cmd_args if a not in addrfiles),
			desc = 'command line',
			edesc = 'argument',
		)

		ad_f = AddrData(self.proto)
		from ..fileutil import check_infile
		for addrfile in addrfiles:
			check_infile(addrfile)
			ad_f.add(AddrList( self.cfg, self.proto, addrfile ))

		ad_w = await TwAddrData(self.cfg,self.proto,twctl=self.twctl)

		await self.process_cmd_args(cmd_args,ad_f,ad_w)

		self.add_mmaddrs_to_outputs(ad_w,ad_f)
		self.check_dup_addrs('outputs')

		if self.chg_output is not None:
			if self.chg_autoselected:
				self.confirm_autoselected_addr(self.chg_output)
			elif len(self.outputs) > 1:
				await self.warn_chg_addr_used(self.chg_output)

	def confirm_autoselected_addr(self,chg):
		from ..ui import keypress_confirm
		if not keypress_confirm(
				self.cfg,
				'Using {a} as {b} address. OK?'.format(
					a = chg.mmid.hl(),
					b = 'single output' if len(self.outputs) == 1 else 'change' ),
				default_yes = True ):
			die(1,'Exiting at user request')

	async def warn_chg_addr_used(self,chg):
		from ..tw.addresses import TwAddresses
		if (await TwAddresses(self.cfg,self.proto,get_data=True)).is_used(chg.addr):
			from ..ui import keypress_confirm
			if not keypress_confirm(
					self.cfg,
					'{a} {b} {c}\n{d}'.format(
						a = yellow('Requested change address'),
						b = (chg.mmid or chg.addr).hl(),
						c = yellow('is already used!'),
						d = yellow('Address reuse harms your privacy and security. Continue anyway? (y/N): ')
					),
					complete_prompt = True,
					default_yes = False ):
				die(1,'Exiting at user request')

	# inputs methods
	def select_unspent(self,unspent):
		prompt = 'Enter a range or space-separated list of outputs to spend: '
		from ..ui import line_input
		while True:
			reply = line_input( self.cfg, prompt ).strip()
			if reply:
				from ..addrlist import AddrIdxList
				selected = get_obj(AddrIdxList, fmt_str=','.join(reply.split()) )
				if selected:
					if selected[-1] <= len(unspent):
						return selected
					msg(f'Unspent output number must be <= {len(unspent)}')

	def select_unspent_cmdline(self,unspent):

		def idx2num(idx):
			uo = unspent[idx]
			mmid_disp = f' ({uo.twmmid})' if uo.twmmid.type == 'mmgen' else ''
			msg(f'Adding input: {idx + 1} {uo.addr}{mmid_disp}')
			return idx + 1

		def get_uo_nums():
			for addr in self.cfg.inputs.split(','):
				if is_mmgen_id(self.proto,addr):
					attr = 'twmmid'
				elif is_coin_addr(self.proto,addr):
					attr = 'addr'
				else:
					die(1,f'{addr!r}: not an MMGen ID or {self.coin} address')

				found = False
				for idx,us in enumerate(unspent):
					if getattr(us,attr) == addr:
						yield idx2num(idx)
						found = True

				if not found:
					die(1,f'{addr!r}: address not found in tracking wallet')

		return set(get_uo_nums()) # silently discard duplicates

	def copy_inputs_from_tw(self,tw_unspent_data):
		def gen_inputs():
			for d in tw_unspent_data:
				i = self.Input(
					self.proto,
					**{attr:getattr(d,attr) for attr in d.__dict__ if attr in self.Input.tw_copy_attrs} )
				if d.twmmid.type == 'mmgen':
					i.mmid = d.twmmid # twmmid -> mmid
				yield i
		self.inputs = type(self.inputs)(self,list(gen_inputs()))

	def warn_insufficient_funds(self,funds_left):
		msg(self.msg_low_coin.format(self.proto.coin_amt(-funds_left).hl(),self.coin))

	async def get_funds_left(self,fee,outputs_sum):
		return self.sum_inputs() - outputs_sum - fee

	async def get_inputs_from_user(self,outputs_sum):

		while True:
			us_f = self.select_unspent_cmdline if self.cfg.inputs else self.select_unspent
			sel_nums = us_f(self.twuo.data)

			msg(f'Selected output{suf(sel_nums)}: {{}}'.format(' '.join(str(n) for n in sel_nums)))
			sel_unspent = MMGenList(self.twuo.data[i-1] for i in sel_nums)

			inputs_sum = sum(s.amt for s in sel_unspent)
			if not await self.precheck_sufficient_funds(inputs_sum,sel_unspent,outputs_sum):
				continue

			self.copy_inputs_from_tw(sel_unspent)  # makes self.inputs

			self.usr_fee = await self.get_fee_from_user()

			funds_left = await self.get_funds_left(self.usr_fee,outputs_sum)

			if funds_left >= 0:
				p = self.final_inputs_ok_msg(funds_left)
				from ..ui import keypress_confirm
				if self.cfg.yes or keypress_confirm( self.cfg, p+'. OK?', default_yes=True ):
					if self.cfg.yes:
						msg(p)
					return funds_left
			else:
				self.warn_insufficient_funds(funds_left)

	async def create(self,cmd_args,locktime=None,do_info=False,caller='txcreate'):

		assert isinstance( locktime, (int,type(None)) ), 'locktime must be of type int'

		from ..tw.unspent import TwUnspentOutputs

		if self.cfg.comment_file:
			self.add_comment(self.cfg.comment_file)

		twuo_addrs = await self.get_input_addrs_from_cmdline()

		self.twuo = await TwUnspentOutputs(self.cfg,self.proto,minconf=self.cfg.minconf,addrs=twuo_addrs)
		await self.twuo.get_data()

		if not do_info:
			await self.get_outputs_from_cmdline(cmd_args)

		from ..ui import do_license_msg
		do_license_msg(self.cfg)

		if not self.cfg.inputs:
			await self.twuo.view_filter_and_sort()

		self.twuo.display_total()

		if do_info:
			del self.twuo.twctl
			import sys
			sys.exit(0)

		outputs_sum = self.sum_outputs()

		msg('Total amount to spend: {}'.format(
			f'{outputs_sum.hl()} {self.dcoin}' if outputs_sum else 'Unknown'
		))

		funds_left = await self.get_inputs_from_user(outputs_sum)

		self.check_non_mmgen_inputs(caller)

		self.update_change_output(funds_left)

		if not self.cfg.yes:
			self.add_comment()  # edits an existing comment

		await self.create_serialized(locktime=locktime) # creates self.txid too

		self.add_timestamp()
		self.add_blockcount()
		self.chain = self.proto.chain_name
		self.check_fee()

		self.cfg._util.qmsg('Transaction successfully created')

		from . import UnsignedTX
		new = UnsignedTX(cfg=self.cfg, data=self.__dict__, automount=self.cfg.autosign)

		if not self.cfg.yes:
			new.info.view_with_prompt('View transaction details?')

		del new.twuo.twctl
		return new



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-txbump: Increase the fee on a replaceable (replace-by-fee) MMGen
              transaction, and optionally sign and send it
"""

from .cfg import gc,Config
from .util import msg,msg_r,die,async_run
from .color import green

opts_data = {
	'sets': [('yes', True, 'quiet', True)],
	'text': {
		'desc': f"""
                Increase the fee on a replaceable (RBF) {gc.proj_name} transaction,
                creating a new transaction, and optionally sign and send the
                new transaction
		 """,
		'usage':   f'[opts] [{gc.proj_name} TX file] [seed source] ...',
		'options': """
-h, --help             Print this help message
--, --longhelp         Print help message for long options (common options)
-a, --autosign         Bump the most recent transaction created and sent with
                       the --autosign option. The removable device is mounted
                       and unmounted automatically.  The transaction file
                       argument must be omitted.  Note that only sent trans-
                       actions may be bumped with this option.  To redo an
                       unsent --autosign transaction, first delete it using
                       ‘mmgen-txsend --abort’ and then create a new one
-b, --brain-params=l,p Use seed length 'l' and hash preset 'p' for
                       brainwallet input
-c, --comment-file=  f Source the transaction's comment from file 'f'
-d, --outdir=        d Specify an alternate directory 'd' for output
-e, --echo-passphrase  Print passphrase to screen when typing it
-f, --fee=           f Transaction fee, as a decimal {cu} amount or as
                       {fu} (an integer followed by {fl!r}).
                       See FEE SPECIFICATION below.
-H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
                      'f' at offset 'o' (comma-separated)
-i, --in-fmt=        f Input is from wallet format 'f' (see FMT CODES below)
-l, --seed-len=      l Specify wallet seed length of 'l' bits. This option
                       is required only for brainwallet and incognito inputs
                       with non-standard (< {dsl}-bit) seed lengths.
-k, --keys-from-file=f Provide additional keys for non-{pnm} addresses
-K, --keygen-backend=n Use backend 'n' for public key generation.  Options
                       for {coin_id}: {kgs}
-M, --mmgen-keys-from-file=f Provide keys for {pnm} addresses in a key-
                       address file (output of '{pnl}-keygen'). Permits
                       online signing without an {pnm} seed source. The
                       key-address file is also used to verify {pnm}-to-{cu}
                       mappings, so the user should record its checksum.
-o, --output-to-reduce=o Deduct the fee from output 'o' (an integer, or 'c'
                       for the transaction's change output, if present)
-O, --old-incog-fmt    Specify old-format incognito input
-p, --hash-preset=   p Use the scrypt hash parameters defined by preset 'p'
                       for password hashing (default: '{gc.dfl_hash_preset}')
-P, --passwd-file=   f Get {pnm} wallet passphrase from file 'f'
-q, --quiet            Suppress warnings; overwrite files without prompting
-s, --send             Sign and send the transaction (the default if seed
                       data is provided)
-v, --verbose          Produce more verbose output
-y, --yes             Answer 'yes' to prompts, suppress non-essential output
-z, --show-hash-presets Show information on available hash presets
""",
	'notes': """
{}{}
Seed source files must have the canonical extensions listed in the 'FileExt'
column below:

FMT CODES:

  {f}
"""
	},
	'code': {
		'options': lambda cfg,help_notes,proto,s: s.format(
			cfg=cfg,
			gc=gc,
			pnm=gc.proj_name,
			pnl=gc.proj_name.lower(),
			fu=help_notes('rel_fee_desc'),
			fl=help_notes('fee_spec_letters'),
			kgs=help_notes('keygen_backends'),
			coin_id=help_notes('coin_id'),
			dsl=help_notes('dfl_seed_len'),
			cu=proto.coin),
		'notes': lambda help_notes,s: s.format(
			help_notes('fee'),
			help_notes('txsign'),
			f=help_notes('fmt_codes')),
	}
}

cfg = Config(opts_data=opts_data)

if not cfg.autosign:
	tx_file = cfg._args.pop(0)
	from .fileutil import check_infile
	check_infile(tx_file)

from .tx import CompletedTX, BumpTX, UnsignedTX, OnlineSignedTX
from .tx.sign import txsign,get_seed_files,get_keyaddrlist,get_keylist

seed_files = get_seed_files(cfg,cfg._args) if (cfg._args or cfg.send) else None

from .ui import do_license_msg
do_license_msg(cfg)

silent = cfg.yes and cfg.fee is not None and cfg.output_to_reduce is not None

async def main():

	if cfg.autosign:
		from .tx.util import init_removable_device
		from .autosign import Signable
		asi = init_removable_device(cfg)
		asi.do_mount()
		si = Signable.automount_transaction(asi)
		if si.unsigned or si.unsent:
			state = 'unsigned' if si.unsigned else 'unsent'
			die(1,
				'Only sent transactions can be bumped with --autosign.  Instead of bumping\n'
				f'your {state} transaction, abort it with ‘mmgen-txsend --abort’ and create\n'
				'a new one.')
		orig_tx = await si.get_last_created()
		kal = kl = sign_and_send = None
	else:
		orig_tx = await CompletedTX(cfg=cfg, filename=tx_file)

	if not silent:
		msg(green('ORIGINAL TRANSACTION'))
		msg(orig_tx.info.format(terse=True))

	if not cfg.autosign:
		kal = get_keyaddrlist(cfg, orig_tx.proto)
		kl = get_keylist(cfg)
		sign_and_send = any([seed_files, kl, kal])

	from .tw.ctl import TwCtl
	tx = await BumpTX(
		cfg  = cfg,
		data = orig_tx.__dict__,
		automount = cfg.autosign,
		check_sent = cfg.autosign or sign_and_send,
		twctl = await TwCtl(cfg,orig_tx.proto) if orig_tx.proto.tokensym else None )

	from .rpc import rpc_init
	tx.rpc = await rpc_init(cfg,tx.proto)

	msg('Creating replacement transaction')

	tx.check_sufficient_funds_for_bump()

	output_idx = tx.choose_output()

	if not silent:
		msg(f'Minimum fee for new transaction: {tx.min_fee.hl()} {tx.proto.coin}')

	tx.usr_fee = tx.get_usr_fee_interactive(fee=cfg.fee,desc='User-selected')

	tx.bump_fee(output_idx,tx.usr_fee)

	assert tx.fee <= tx.proto.max_tx_fee

	if not cfg.yes:
		tx.add_comment()   # edits an existing comment

	await tx.create_serialized(bump=True)

	tx.add_timestamp()
	tx.add_blockcount()

	cfg._util.qmsg('Fee successfully increased')

	if not silent:
		msg(green('\nREPLACEMENT TRANSACTION:'))
		msg_r(tx.info.format(terse=True))

	if sign_and_send:
		tx2 = UnsignedTX(cfg=cfg,data=tx.__dict__)
		tx3 = await txsign(cfg,tx2,seed_files,kl,kal)
		if tx3:
			tx4 = await OnlineSignedTX(cfg=cfg,data=tx3.__dict__)
			tx4.file.write(ask_write=False)
			if await tx4.send():
				tx4.file.write(ask_write=False)
		else:
			die(2,'Transaction could not be signed')
	else:
		tx.file.write(
			outdir                = asi.txauto_dir if cfg.autosign else None,
			ask_write             = not cfg.yes,
			ask_write_default_yes = False,
			ask_overwrite         = not cfg.yes)

async_run(main())
